"""## QFL Experiment Parallel Function"""

import copy
import math
import pickle
from pennylane import numpy as np
import torch

from quorus.data_ops.data_loading_wrapper import load_dataset
from quorus.logging.custom_slog import print_cust
from quorus.metrics_funcs.agg_metrics_func.agg_metrics_func_eval import compute_metrics_angle_param_batch
from quorus.metrics_funcs.fid_funcs.compute_fid import compute_fid_to_data
from quorus.misc_utils.convert_data_to_lib import convert_data_to_lib
from quorus.misc_utils.create_feat_list_expansion_data import create_feat_list_expansion_data
from quorus.misc_utils.reorder_amplitude_data import reorder_amplitude_data
from quorus.qfl_utils.aggregation_funcs.federated_averaging import federated_averaging
from quorus.qfl_utils.aggregation_funcs.federated_averaging_circular_parallel_shared import federated_averaging_circular_parallel_shared
from quorus.qfl_utils.aggregation_funcs.federated_averaging_quat import federated_averaging_quat
from quorus.qfl_utils.broadcast_params_funcs.broadcast_param_updates import broadcast_param_updates
from quorus.qfl_utils.broadcast_params_funcs.broadcast_param_updates_shared import broadcast_param_updates_shared
from quorus.qfl_utils.broadcast_params_funcs.broadcast_param_updates_shared_noconv import broadcast_param_updates_shared_noconv
from quorus.qfl_utils.broadcast_params_funcs.broadcast_param_updates_shared_noconv_nofinal import broadcast_param_updates_shared_noconv_nofinal
from quorus.qfl_utils.cli_params_init.initialize_client_params import initialize_client_params
from quorus.qfl_utils.data_splitting.split_data_federated import split_data_federated
from quorus.qfl_utils.federated_pca.fedpca_mocked import perform_federated_pca_mocked
from quorus.qfl_utils.meta_params_generation.generate_meta_params import generate_meta_params
from quorus.qfl_utils.meta_params_generation.generate_meta_params_random import generate_meta_params_random
from quorus.qgan_model_supp.imggen_funcs.latent_noise_gen import generate_latent_noise
import torch.nn as nn

from quorus.qgan_model_supp.imggen_funcs.save_imgs import save_tensors_to_folder
from quorus.qnode_funcs.qnode_creation_funcs.create_qnode_qcnn import create_qnode_qcnn
from quorus.qnode_funcs.qnode_creation_funcs.create_qnode_qgan import create_qnode_qgan
from quorus.quantum_circuit_funcs.utils.qcnn_code import compute_conv_layers
from quorus.training_funcs.pennylane_training.mask_gradients import mask_gradients
from quorus.training_funcs.pennylane_training.train_mod_pennylane import train_epochs_angle_param_adam
from quorus.training_funcs.torch_training.qgan_training.qgan_training_funcs import train_models

def run_qfl_experiments_parallel(clients_config_arg, classes=["4", "9"], n_samples=1000, dataset_type="mnist", agg_strategy="fedavg", test_frac=0.2, val_frac=0.1, random_state=42, pool_in=True,
                        local_batch_size=32, local_lr=0.01, shots=1024, debug=False, init_client_data_dict=None, save_pkl=False, mask_grads=False, qubits_and_layers_to_add_block_params=[],
                        train_models_parallel=False, same_init=False, feature_skew=0.0, label_skew=None, local_pca=False, do_lda=False, feat_sel_type="top", amp_embed=False, feat_ordering="same",
                                 morepers=False, custom_debug=False, shared_pca=False, heirarchical_train=False, generative=False, use_torch=False, fed_pca_mocked=True, lr_gen=0.004, lr_disc=0.001,
                                 noise_func=generate_latent_noise, criterion_func=nn.BCELoss, targ_data_folder_prefix="testing_gen_imgs", gen_data_folder_prefix="qgan_gen_imgs", device=None, fid_batch_size=None):

  """
  Function that runs the QFL workflow for a given configuration and returns a log of the training.

  Note: in general, the client parameters dictionary is of the form:

      {
        <int>: [[conv_params, pool_params, final_pool_params,
            final_params, bias_param, block_params],
            ... for each client],
        ... for each client type
      }

  Parameters:
    clients_config_arg, a dictionary mapping client types to configuration information for clients of that type. For this code, I expect clients_config_arg to be of the form:
      {
        <int>: {
            "percentage_data": <float>,
            "num_clients": <int>,
            "local_epochs": <int>,
            "communication_rounds": <int>
        },
        ... for each client type
      }
    classes, a list of strings (preferably string representations of integers) representing the classes on which to perform classification
      ex: ["4", "9"]
    n_samples, an integer representing the number of samples to use for training
    dataset_type, a string representing the dataset type for training
      should either be "mnist", "Fashion-MNIST", "cifar10", "synthetic", "pima", "higgs", or "covertype" for those respective datasets
    agg_strategy, a string representing the aggregation scheme to perform
      should be either "fedavg", "fedavg_quat", or "fedavg_circ"
    test_frac, a float representing the fraction of the total data to be used for testing
      should be between 0 and 1
    val_frac, a float representing the fraction of each client's data to be used for validation
      should be between 0 and 1
    random_state, an integer representing the random state used in the entire program (assuming that this function generates client data -- TODO: make the use of the random
    state argument more clear)
    pool_in, a Boolean representing whether or not the model should have reduce the number of qubits to the center
    local_batch_size, an integer representing the batch size each local model should use for training
    local_lr, a float representing the learning rate each model should use for training, upon each communication round
    shots, an integer representing the number of shots each model should use (currently unused)
    debug, a Boolean indicating whether or not this function should be run in debug mode (prints the data, the amount of data per client)
    init_client_data_dict, a dictionary that maps client types to lists of data for each client that is used for each client
      should be of the form
      {
        <int>: [[(X_train, y_train), (X_val, y_val), (pca_obj, pca_reduced_data)],
                ... for each client
              ],
        ... for each client type
      }
    save_pkl, a Boolean indicating whether or not the data should be saved to a pickle file
    mask_grads, a Boolean indicating whether or not the gradients should have a Boolean mask (and parameters in smaller qubit sets should not be updated)
    qubits_and_layers_to_add_block_params, a dictionary that maps client types to a list of (n_qubits, n_layers) representing the block parameters that each client has
    train_models_parallel, a Boolean indicating whether or not the models should be trained in parallel (for personalized models)
    same_init, a Boolean indicating whether or not all the clients should start with the same initial parameters
    feature_skew, a float between 0 and 1 specifying the magnitude of the feature skew (i.e., the strength of sorting the features by the first feature and how prominently that appears
    in the data for clients)
    label_skew, a float between 0 and 1 specifying the magnitude of the label skew that each client has
    do_lda, a Boolean indicating whether or not to perform random sketching
    feat_sel_type, a string representing the choice of features to make in angle encoding for each client
      should be either "top" for selecting the features with the highest variance, or "toplow" for selecting half the features with the highest variance and half the features with the
      lowest variance at each expansion
    amp_embed, a Boolean indicating whether the data should be amplitude encoded
    feat_ordering, a string representing whether the features in amplitude encoding should be taken as-is, or if it should be taken in a different order
      should be either "same" for as-is feature ordering, or "highest_var" for sorting the features in descending order in terms of highest variance
    morepers, a Boolean indicating whether or not only the the convolutional parameters should or should not be aggregated (if not, the convolutional parameters are unique to each client,
    so each client has its own personalized convolutional parameters -- see PDF drawing sent in Slack)

    Returns:
      data_logs, a dictionary of the following form:
        {
          'clients_data_dict': <client_data_dict> of the above form,
          'testing_data': (X_test, y_test),
          0:
            {
              "aggregated_params": <params>,
              <int>: {
                "local_epochs": <int>,
                "client_metrics": [
                    {
                      "trained_params": <params>,
                      "minibatch_losses": list<float>,
                      "validation_losses": list<float>,
                      "training_acc": <float>,
                      "testing_acc": <float>,
                      "testing_loss": <float>,
                      "training_acc_stdev": <float>,
                      "testing_acc_stdev": <float>,
                      "training_acc_topk": <float>,
                      "testing_acc_topk": <float>
                    }, ... (for each client)
                ]
              },
              ... for each client type
            },
          ... for each communication round
          clients_config_arg["communication_rounds"] - 1: (same format as above data logs dictionary)
        }
  """

  # TOADD, generative:
  # 1. An option to this function to indicate that we are having a generative model

  print_cust(f"run_qfl_experiments_parallel, heirarchical_train: {heirarchical_train}")
  # Find the maximum, minimum size clients, as well as the total communication rounds
  max_size_clients = max(clients_config_arg.keys())
  min_size_clients = min(clients_config_arg.keys())

  if custom_debug:
    print_cust(f"run_qfl_experiments_parallel, max_size_clients: {max_size_clients}, min_size_clients, {min_size_clients}")
    assert max_size_clients >= min_size_clients, "Maximum sized client is NOT at least as large as minimum sized client"

  num_total_rounds = max([clients_config_arg[key]["communication_rounds"] for key in clients_config_arg])

  # Note: this does impose some more constraints on the format in which the round info is passed (larger MUST be > smaller; training 4 implies training 8.) going to stick with this for now
  total_rounds_accum = 0
  client_types_to_rounds = {}
  for client_type in sorted(clients_config_arg.keys()):
    client_rounds = clients_config_arg[client_type]["communication_rounds"]
    client_types_to_rounds[client_type] = client_rounds - total_rounds_accum
    total_rounds_accum += client_rounds

  # this should be client type to rounds TO RUN for basically clients that size AND ABOVE.
  print_cust(f"run_qfl_experiments_parallel, client_types_to_rounds: {client_types_to_rounds}")

  if custom_debug:
    print_cust(f"run_qfl_experiments_parallel, num_total_rounds: {num_total_rounds}")
    assert num_total_rounds >= 0, "Number of total rounds is not at least 0"

  # The number of output qubits is enough to accommodate the number of classes.
  n_output_qubits = int(np.ceil(np.log2(len(classes))))

  if custom_debug:
    print_cust(f"run_qfl_experiments_parallel, n_output_qubits: {n_output_qubits}")
    assert n_output_qubits >= 0, "Number of output qubits is not at least 0"

  print_cust(f"run_qfl_experiments_parallel, num_total_rounds: {num_total_rounds}")

  print_cust(f"run_qfl_experiments_parallel, n_output_qubits: {n_output_qubits}")

  # Load the input dataset. If we are doing amplitude embedding or local PCA, then we do not want to dimensionality reduce the input images.
  # Note that the pixels are normalized to be between 0 and 1.
  keep_orig_imgs = (local_pca or amp_embed)

  if custom_debug:
    print_cust(f"run_qfl_experiments_parallel, keep_orig_imgs: {keep_orig_imgs}")

  X_angles, y = load_dataset(dataset_type=dataset_type, classes=classes, n_samples=n_samples, num_feats=max_size_clients, keep_orig_imgs=keep_orig_imgs, custom_debug=custom_debug)

  print_cust(f"run_qfl_experiments_parallel, X_angles.shape: {X_angles.shape}")

  print_cust(f"run_qfl_experiments_parallel, local_pca: {local_pca}, do_lda: {do_lda}")

  # Load in the previous data dictionary, if supplied.
  if init_client_data_dict is not None:
    clients_data_dict = init_client_data_dict["clients_data_dict"]
    (X_test, y_test) = init_client_data_dict["testing_data"]
    # if shared_pca:
    #   shared_max_comps = init_client_data_dict['shared_max_comps']
    #   shared_min_comps = init_client_data_dict['shared_min_comps']
    print_cust(f"run_qfl_experiments_parallel, loaded in existing data")
  else:
    clients_data_dict, (X_test, y_test) = split_data_federated(X_angles, y, clients_config_arg, test_frac, val_frac=val_frac, random_state=random_state, feature_skew=feature_skew, label_skew=label_skew, local_pca=local_pca,
                                                              do_lda=do_lda, feat_sel_type=feat_sel_type, amp_embed=amp_embed, feat_ordering=feat_ordering, shared_pca=shared_pca, fed_pca_mocked=fed_pca_mocked)
    print_cust(f"run_qfl_experiments_parallel, generated new data")

  # Create a log of the information that we have throughout training.
  data_logs = {}

  data_logs["clients_data_dict"] = clients_data_dict
  data_logs["testing_data"] = (X_test, y_test)
  # if shared_pca:
  #   data_logs['shared_max_comps'] = shared_max_comps
  #   data_logs['shared_min_comps'] = shared_min_comps

  # Store the initial set of training parameters and data in a pickle file.
  if save_pkl:
    with open(f"data_logs_n_samples_{n_samples}_dataset_type_{dataset_type}_classes_{'_'.join(classes)}_train_models_parallel_{train_models_parallel}_feature_skew_{feature_skew}_label_skew_{label_skew}_local_pca_{local_pca}_shared_pca_{shared_pca}_gen_{generative}.pkl", "wb") as file:
      pickle.dump(data_logs, file)


  if use_torch:
    math_int = torch
  else:
    math_int = np

  # Maybe right here, have a torch data conversion function ??? (if not done in perform_federated_pca_mocked)
  clients_data_dict = convert_data_to_lib(clients_data_dict, math_int=math_int)
  if math_int == torch:
    X_test = torch.tensor(X_test, dtype=torch.float32)

  if generative:
    print_cust(f"run_qfl_experiments_parallel, X_test.shape: {X_test.shape}")
    if len(X_test.shape) == 2:
      img_dim = math.isqrt(X_test.shape[1])
      assert (img_dim ** 2) == X_test.shape[1], f"run_qfl_experiments_parallel, X_test is not a perfect square, X_test.shape: {X_test.shape}"
      X_test = X_test.view(X_test.shape[0], img_dim, img_dim)
    print_cust(f"run_qfl_experiments_parallel, generative, X_test.max(): {X_test.max()}, X_test.min(): {X_test.min()}")
    save_tensors_to_folder(X_test, f"{targ_data_folder_prefix}", "img")
    n_samples_test = X_test.shape[0]


  # TODO: continue reading here, 7/9
  # also, please put this in a function......
  if shared_pca:
    # NOTE: this function MUTATES clients_data_dict
    pca_global, shared_min_comps, shared_max_comps, inv_pca_min, inv_pca_max = perform_federated_pca_mocked(clients_data_dict, max_size_clients, random_seed=random_state, math_int=math_int, device=device)
    # TODO: ... = perform_federated_pca_mocked()...

  # For sanity, print out the data as well as the amount of data that each client has.
  if debug:
    for client_type in clients_data_dict:
      clients_data_list = clients_data_dict[client_type]
      print_cust(f"client_type: {client_type}, len(clients_data_list): {len(clients_data_list)}")
      for client_idx in range(len(clients_data_list)):
        client_data = clients_data_list[client_idx]
        print_cust(f"client_type: {client_type}, client_idx: {client_idx}, len(client_data): {len(client_data)}")
        if not local_pca:
          (training_data, validation_data) = client_data
        else:
          (training_data, validation_data, pca_info) = client_data
        print_cust(f"client_type: {client_type}, client_idx: {client_idx}, len(training_data): {len(training_data)}, len(validation_data): {len(validation_data)}")
        (X_train, y_train) = training_data
        (X_val, y_val) = validation_data
        print_cust(f"client_type: {client_type}, client_idx: {client_idx}, X_train.shape: {X_train.shape}, y_train.shape: {y_train.shape}, X_val.shape: {X_val.shape}, y_val.shape: {y_val.shape}")

  # Initialize the parameters for each client.
  # qnode_func, device, pca_info
  qnode_func = None
  # device = None
  pca_info = ()
  if generative:
    qnode_func = create_qnode_qgan
    # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    pca_info = (pca_global, shared_min_comps, shared_max_comps, inv_pca_min, inv_pca_max)
  client_params_dict = initialize_client_params(clients_config_arg, model_size=min_size_clients, cur_client_params_dict=None, qubits_and_layers_to_add_block_params=qubits_and_layers_to_add_block_params, train_models_parallel=train_models_parallel,
                                                n_output_qubits=n_output_qubits, generative=generative, use_torch=use_torch, qnode_func=qnode_func, device=device, pca_info=pca_info)
  print_cust(f"run_qfl_experiments_parallel, first initialization of client params, client_params_dict: {client_params_dict}")

  # Make all parameters the same across clients with a new random initialization if we want clients to have the same initialization.
  if same_init:
    print_cust(f"run_qfl_experiments_parallel, same parameters initialization")
    with torch.no_grad():
      meta_params = generate_meta_params(client_params_dict, clients_data_dict, math_int=math_int)
      meta_params = generate_meta_params_random(meta_params, math_int=math_int)
      client_params_dict = broadcast_param_updates(client_params_dict, meta_params)
      print_cust(f"run_qfl_experiments_parallel, client_params_dict: {client_params_dict}")

  # client_types = sorted(list(clients_config_arg.keys()))

  # Initialize the data logging information for each communication round.

  for round_num in range(num_total_rounds):
    print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}")
    # Initialize the data logging information for this communication round.
    data_logs[round_num] = {}
    data_logs[round_num]["aggregated_params"] = None
    for client_size, cfg in clients_config_arg.items():
      data_logs[round_num][client_size] = {}
      # TODO: add a field for testing loss in data_logs
      data_logs[round_num][client_size]["local_epochs"] = cfg["local_epochs"]
      data_logs[round_num][client_size]["client_metrics"] = []
      for client_idx in range(cfg["num_clients"]):
        data_logs[round_num][client_size]["client_metrics"].append({"trained_params": None,
                                                      "minibatch_losses": None,
                                                      "validation_losses": None,
                                                      "training_acc": None,
                                                      "testing_acc": None,
                                                      "testing_loss": None,
                                                      "training_acc_stdev": None,
                                                      "testing_acc_stdev": None,
                                                      "training_acc_topk": None,
                                                      "testing_acc_topk": None,
                                                      "testing_fid": None})

  rounds_elapsed = 0
  for cur_shared_model_size, model_size_rounds in client_types_to_rounds.items():

    if cur_shared_model_size > min_size_clients and not train_models_parallel:
      # two things to check:
      # 1. initialization with existing parameters - OK
      # 2. expansion from smaller to larger qubits. - easy fix in pca_rescaler
      client_params_dict = initialize_client_params(clients_config_arg, model_size=cur_shared_model_size, cur_client_params_dict=client_params_dict, qubits_and_layers_to_add_block_params=qubits_and_layers_to_add_block_params, train_models_parallel=train_models_parallel,
                                                n_output_qubits=n_output_qubits, generative=generative, use_torch=use_torch, qnode_func=qnode_func, device=device, pca_info=pca_info)
      print_cust(f"run_qfl_experiments_parallel, after expansion, client_params_dict: {client_params_dict}")
      existing_clients_datadict = list(clients_data_dict.keys())
      for existing_client_size in existing_clients_datadict:
        if cur_shared_model_size > existing_client_size:
          del clients_data_dict[existing_client_size]

      print_cust(f"run_qfl_experiments_parallel, not train_models_parallel, clients_data_dict.keys(): {clients_data_dict.keys()}")

    # For each communication round, iterate over each client and have each client train on their dataset for the specified number of epochs.
    for round_num in range(model_size_rounds):
      round_num += rounds_elapsed
      print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}")

      for client_type, client_params in client_params_dict.items():
        # TODO: add a condition, that, based on the current model size, if heirarchical, skips particular client types.
        print_cust(f"run_qfl_experiments_parallel, client_type: {client_type}")
        # Get the clients' data and the number of convolutional layers for clients of this type.
        client_data = clients_data_dict[client_type]
        if train_models_parallel:
          cur_model_size = client_type
        else:
          cur_model_size = cur_shared_model_size
        print_cust(f"run_qfl_experiments_parallel, cur_model_size: {cur_model_size}")
        conv_layers = compute_conv_layers(cur_model_size, n_output_qubits, generative=generative)
        print_cust(f"run_qfl_experiments_parallel, conv_layers: {conv_layers}")
        for client_idx in range(len(client_params)):
          print_cust(f"run_qfl_experiments_parallel, client_idx: {client_idx}")
          # Get the training data and parameters for each client.
          client_params_indiv = client_params[client_idx]
          client_data_indiv = client_data[client_idx]

          client_train_data = client_data_indiv[0]
          client_val_data = client_data_indiv[1]

          if debug:
            print_cust(f"run_qfl_experiments_parallel, len(client_train_data): {len(client_train_data)}")
            print_cust(f"run_qfl_experiments_parallel, client_train_data: {client_train_data}")
            print_cust(f"run_qfl_experiments_parallel, len(client_val_data): {len(client_val_data)}")
            print_cust(f"run_qfl_experiments_parallel, client_val_data: {client_val_data}")

          grad_mask = None

          # Gradient masking hasn't been tested for a while, so currently is commented out.
          # TODO: implement gradient masking!!!!!!

          if mask_grads:
            if cur_model_size > min_size_clients and not generative:
              grad_mask = mask_gradients(client_params_indiv)

          print_cust(f"run_qfl_experiments_parallel, grad_mask: {grad_mask}")

          # Create the feature list indicating what features for which qubits are to be used for this client, as well as the expansion data (used for identity
          # initialization).
          expand = False
          if cur_model_size > min_size_clients:
            expand = True
          if not generative:
            feature_list, expansion_data = create_feat_list_expansion_data(cur_model_size, conv_layers, expand=expand, pool_in=pool_in, min_qubits_noexpand=min_size_clients,
                                                                          train_models_parallel=train_models_parallel, feat_sel_type=feat_sel_type)
          else:
            feature_list, expansion_data = list(range(cur_model_size)), []

          # Create the QNode for this client.
          # NOTE: for generative, there is no need to create a separate qnode; it should already be created for the generative model, upon instantiation.
          if not generative:
            qnode = create_qnode_qcnn(cur_model_size, conv_layers, expansion_data, n_classes=len(classes))

          num_local_epochs = clients_config_arg[client_type]["local_epochs"]

          if not generative:
            client_params_indiv = copy.deepcopy(client_params_indiv)

          # If amplitude embedding the data, reorder the data; otherwise, sample features based on feature_list.
          if amp_embed:
            training_data = reorder_amplitude_data(client_train_data[0], feature_list)
            validation_data = reorder_amplitude_data(client_val_data[0], feature_list)
          else:
            training_data = client_train_data[0][:, feature_list]
            validation_data = client_val_data[0][:, feature_list]

          # Train the client for the specified number of epochs.
          # TODO: update the grad_mask here.
          if not generative:
            trained_params, minibatch_losses, validation_losses = train_epochs_angle_param_adam(
          client_params_indiv, training_data, client_train_data[1], validation_data, client_val_data[1],
          n_epochs=num_local_epochs, shots=shots, batch_size=local_batch_size, lr=local_lr, qnode=qnode, trainable_mask=grad_mask)
            # Update the client's trained parameters.
            client_params[client_idx] = trained_params
          else:
            # I mean, I could return the minibatch_losses, validation_losses, but idt I will.
            print_cust(f"run_qfl_experiments_parallel, generative is True (QGAN training)")
            client_generator = client_params_indiv[5][0]
            print_cust(f"run_qfl_experiments_parallel, type(client_generator): {type(client_generator)}")
            client_discriminator = client_params_indiv[5][1]
            print_cust(f"run_qfl_experiments_parallel, type(client_discriminator): {type(client_discriminator)}")
            print_cust(f"run_qfl_experiments_parallel, nn.ParameterList([client_generator.q_params[-1]]: {nn.ParameterList([client_generator.q_params[-1]])}")
            client_optim_gen = torch.optim.SGD(nn.ParameterList([client_generator.q_params[-1]]), lr=lr_gen)
            client_optim_disc = torch.optim.SGD(client_discriminator.parameters(), lr=lr_disc)
            # maybe concatenate in the train with val data? not using val data otherwise
            # can use test_result_imgs for visualization, later, if I'd like.
            test_result_imgs = train_models(client_generator.n_qubits_gen, local_batch_size, client_generator, client_discriminator, client_optim_disc, client_optim_gen,
                                            noise_func=noise_func, criterion=criterion_func(), train_data=training_data, device=device, image_size=0, compressed_img_size=0, max_num_epochs=num_local_epochs,
                                            n_qubits_small=0, gen_pcas=True, disc_img_size=None, pca_disc=True)

          # Perform PCA on the testing data, if specified.
          if local_pca and not generative:
            client_pca_info = client_data_indiv[2]
            client_pca = client_pca_info[0]
            client_data_pca = client_pca_info[1]
            X_test_client_pca = client_pca.transform(X_test)
            # Scale each PCA component independently to [0, π]
            X_test_client_angle = math_int.zeros_like(X_test_client_pca)
            # assuming that the number of components is simply the client's data type
            # TODO: rescale the testing data to match the PCA scale for each client
            for i in range(client_type):
                comp = X_test_client_pca[:, i]
                if shared_pca:
                  lo, hi = shared_min_comps[i], shared_max_comps[i]
                  comp_norm = ( (comp - lo) / (hi - lo + 1e-8) )
                  # comp_norm = np.clip(comp_norm, 0, 1)
                else:
                  orig_comp = client_data_pca[:, i]
                  comp_norm = (comp - orig_comp.min()) / (orig_comp.max() - orig_comp.min() + 1e-8)
                X_test_client_angle[:, i] = comp_norm * math_int.pi
          else:
            X_test_client_angle = X_test

          # Select the features for amplitude encoding based on feat_ordering.
          if amp_embed:
            if feat_ordering == "highest_var":
              variances = X_angles.var(axis=0, ddof=0)
              order = np.argsort(variances)[::-1]
              X_test_client_angle = X_test_client_angle[:, order] + 1e-3
            testing_data = reorder_amplitude_data(X_test_client_angle, feature_list)
          else:
            if not generative:
              testing_data = X_test_client_angle[:, feature_list]

            # TODO: encapsulate this logic in a helper function


          # train_acc = compute_avg_acc_angle_param_batch(trained_params, client_train_data[0][:, feature_list], client_train_data[1], layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

          # test_acc = compute_avg_acc_angle_param_batch(trained_params, X_test_client_angle[:, feature_list], y_test, layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

          # test_loss = compute_loss_angle_param_batch(trained_params, X_test_client_angle[:, feature_list], y_test, shots=shots, qnode=qnode)

          # train_acc_stdev = compute_std_acc_angle_param_batch(trained_params, client_train_data[0][:, feature_list], client_train_data[1], layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

          # test_acc_stdev = compute_std_acc_angle_param_batch(trained_params, X_test_client_angle[:, feature_list], y_test, layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

          # train_acc_topk = compute_top_k_acc_angle_param_batch(trained_params, client_train_data[0][:, feature_list], client_train_data[1], layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

          # test_acc_topk = compute_top_k_acc_angle_param_batch(trained_params, X_test_client_angle[:, feature_list], y_test, layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

          # Compute the training and testing metrics.
          if not generative:
            train_acc, train_acc_stdev, train_acc_topk, train_loss = compute_metrics_angle_param_batch(trained_params, training_data, client_train_data[1], layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)

            # Store the training and testing metrics.
            data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["training_acc"] = train_acc
            # data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["testing_acc"] = test_acc
            # data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["testing_loss"] = test_loss

            data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["training_acc_stdev"] = train_acc_stdev
            # data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["testing_acc_stdev"] = test_acc_stdev

            data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["training_acc_topk"] = train_acc_topk
            # data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["testing_acc_topk"] = test_acc_topk

            data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["trained_params"] = trained_params
            data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["minibatch_losses"] = minibatch_losses
            data_logs[round_num][cur_model_size]["client_metrics"][client_idx]["validation_losses"] = validation_losses

      # perform aggregation
      # NOTE: assume, during aggregation, that the parameters for all the clients have the same shape.
      # Aggregate parameters based on the strategy provided.
      print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}, parameter aggregation, agg_strategy: {agg_strategy}")

      print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}, before aggregation, client_params_dict: {client_params_dict}")

      if agg_strategy == "fedavg":
        aggregated_params = federated_averaging(client_params_dict, clients_data_dict)
      elif agg_strategy == "fedavg_quat":
        aggregated_params = federated_averaging_quat(client_params_dict, clients_data_dict)
      elif agg_strategy == "fedavg_circ":
        aggregated_params = federated_averaging_circular_parallel_shared(client_params_dict, clients_data_dict, generative=generative)
        # print_cust(f"run_qfl_experiments, aggregated_params: {aggregated_params}")

      print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}, after aggregation, aggregated_params: {aggregated_params}")

      # # broadcast parameter updates
      # for client_type, client_params in client_params_dict.items():
      #   for client_idx in range(len(client_params)):
      #     client_params[client_idx] = aggregated_params

      # client_params_dict = broadcast_param_updates_shared(client_params_dict, aggregated_params)

      # Broadcast parameters based on the personalization scheme.
      if morepers == "aggshared":
        client_params_dict = broadcast_param_updates_shared(client_params_dict, aggregated_params)
      elif morepers == "aggnoconv":
        client_params_dict = broadcast_param_updates_shared_noconv(client_params_dict, aggregated_params)
      elif morepers == "aggnoconvnofinal":
        client_params_dict = broadcast_param_updates_shared_noconv_nofinal(client_params_dict, aggregated_params)

      print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}, after broadcasting, client_params_dict: {client_params_dict}")

      # Store the aggregated parameters for this round.
      # TODO: aggregated parameters don't have a significant meaning as not all of them are broadcasted to clients based on the broadcasting/personalization scheme
      data_logs[round_num]["aggregated_params"] = aggregated_params


      # TODO: recompute metrics of the SHARED model for EACH client.
      one_cli_size = list(client_params_dict.keys())[0]

      if generative:
        # NOTE: this assumes that all generators and discriminators have the same parameters.
        agg_model = client_params_dict[one_cli_size][0][5][0]
        # agg_disc = client_params_dict[one_cli_size][0][5][1]
      else:
        agg_model = client_params_dict[one_cli_size][0]


      test_acc, test_acc_stdev, test_acc_topk, test_loss, test_fid_score = (None, None, None, None, None)

      if fid_batch_size is None: # for now, for debug
        fid_batch_size = n_samples_test

      if num_total_rounds > 10:
        if (round_num + 1) % 10 == 0:
          if not generative:
            # TODO: this is broken for the non-generative case. process the testing data outside of the loop for each client.. not doing that now tho; not too relevant for
            # generative case
            test_acc, test_acc_stdev, test_acc_topk, test_loss = compute_metrics_angle_param_batch(agg_model, testing_data, y_test, layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)
          else:
            folder_id_suffix = f"round_num_{round_num}_cur_shared_model_size_{cur_shared_model_size}"
            test_fid_score = compute_fid_to_data(agg_model, noise_func, f"{targ_data_folder_prefix}",
                                                f"{gen_data_folder_prefix}/{folder_id_suffix}", n_samples_test, device, fid_batch_size=fid_batch_size)
        else:
          test_acc, test_acc_stdev, test_acc_topk, test_loss = (None, None, None, None)
      else:
        if not generative:
          test_acc, test_acc_stdev, test_acc_topk, test_loss = compute_metrics_angle_param_batch(agg_model, testing_data, y_test, layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)
        else:
          folder_id_suffix = f"round_num_{round_num}_cur_shared_model_size_{cur_shared_model_size}"
          test_fid_score = compute_fid_to_data(agg_model, noise_func, f"{targ_data_folder_prefix}",
                                              f"{gen_data_folder_prefix}/{folder_id_suffix}", n_samples_test, device, fid_batch_size=fid_batch_size)
      if not generative:
        print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}, test_acc: {test_acc}, test_loss: {test_loss}, test_acc_stdev: {test_acc_stdev}, test_acc_topk: {test_acc_topk}")
      else:
        print_cust(f"run_qfl_experiments_parallel, round_num: {round_num}, cur_model_size: {cur_model_size}, test_fid_score: {test_fid_score}")

    rounds_elapsed += model_size_rounds

  return data_logs


"""## Core training function"""

import copy
from pennylane.optimize import AdamOptimizer
from quorus.logging.custom_slog import print_cust
from pennylane import numpy as np
import pennylane as qml
from quorus.metrics_funcs.pennylane_lossfns.pennylane_lossfn_batch import compute_loss_angle_param_batch
from quorus.param_processing.numpy_funcs.param_grads_utils_pennylane import tree_to_list, grad_l2_norm
from quorus.param_processing.numpy_funcs.param_comparison import tuples_allclose

# NOTE: this is NOT used in final quorus version; is here because previous iterations of this project used this func.
# so, no guarantees that this function SPECIFICALLY works as intended (particularly, if autograd/backprop yield the exact,
# expected, correct gradients)
def train_epochs_angle_param_adam(params, X_angles, y, X_val, y_val, n_epochs=5, layers=1, shots=1024, batch_size=32,
                                  lr=0.01, beta1=0.9, beta2=0.999, epsilon=1e-8,
                                  grad_func=None, trainable_mask=None, qnode=None,
                                  patience=3):

    """
    Trains through n_epochs epochs over the given X_angles dataset in minibatches, using the validation set to have early termination.

    Parameters:
      params: a list of numpy arrays
      X_angles: Array-like, inputs to the qnode function.
      y: Array-like, true labels.
      X_val: Array-like, inputs to qnode function used for early termination.
      y_val: Array-like, true labels used for early termination.
      n_epochs: Integer, number of epochs used for training.
      layers: (Optional) Parameter for potential layer configuration (currently unused).
      shots: (Optional) Parameter for simulation shots (currently unused).
      batch_size: (Optional) Parameter for batch processing (currently unused).
      lr: Integer representing the learning rate for the optimizer.
      beta1: Integer representing the beta1 parameter for the optimizer (currently unused).
      beta2: Integer representing the beta2 parameter for the optimizer (currently unused).
      epsilon: Float representing some minimum precision (currently unused).
      grad_func: Function used to apply the gradient (currently unused; could be used in the future for more decoupling/separation of concerns).
      trainable_mask: List of numpy arrays used to specify what parameters to NOT apply the gradient on.
      qnode: A function that takes (input_angles, params) and returns a vector of probabilities.
      patience: Integer representing the maximum number of epochs that is allowed for validation loss to not increase.

    Returns:
      params: a list of numpy arrays representing the trained params
      minibatch_losses: a numpy array containing the losses for each minibatch
      validation_losses: a numpy array containing the losses on the validation set
    """
    # Initialize containers to store metrics as well as the optimizer
    n_samples = X_angles.shape[0]
    minibatch_losses = []
    validation_losses = []
    opt = AdamOptimizer(stepsize=lr)
    best_val_loss = float('inf')
    patience_count = 0
    # print_cust(f"train_epochs_angle_param_adam, params[0].shape: {params[0].shape}")

    # Record the original nested structure as a template
    structure_template = params

    print_cust(f"train_epochs_angle_param_adam, n_samples: {n_samples}")
    # For each epoch, train on randomly sampled minibatches.
    for epoch in range(n_epochs):
        print_cust(f"[Adam SGD] Epoch {epoch+1}/{n_epochs}")
        indices = np.arange(n_samples)
        np.random.shuffle(indices)
        for start in range(0, n_samples, batch_size):
            end = min(start + batch_size, n_samples)
            batch_indices = indices[start:end]
            X_batch = X_angles[batch_indices]
            y_batch = y[batch_indices]
            old_params = copy.deepcopy(params)

            # Compute gradient on the mini-batch
            grads = qml.grad(compute_loss_angle_param_batch, argnum=0)(
                params, X_batch, y_batch, layers=layers, shots=shots, batch_size=len(X_batch), qnode=qnode
            )

            # --------- flatten *one* level ----------
            flat_p,  rebuild = tree_to_list(params)
            flat_g, _        = tree_to_list(grads)       # identical structure ⇒ same rebuild

            # ensure differentiable type
            flat_p = [np.asarray(p, requires_grad=True) for p in flat_p]
            # print_cust(f"train_epochs_angle_param_adam, grads after grad calculation: {grads}")
            # grads = np.array(grads).flatten()
            # If a trainable mask is provided, apply it elementwise
            # if trainable_mask is not None:
            #     def apply_mask(grad, mask):
            #         """Element‑wise multiply `grad` by `mask`, preserving the nested structure.

            #         Both grad and mask must have *exactly* the same tree layout.
            #         Mask entries can be 0/1, booleans, or any broadcast‑compatible scalars.
            #         """
            #         # Same container (list vs tuple) is preserved with `type(grad)(...)`
            #         if isinstance(grad, (list, tuple)):
            #             if len(grad) != len(mask):
            #                 raise ValueError("Mask and gradient have different lengths")
            #             return type(grad)(apply_mask(g, m) for g, m in zip(grad, mask))
            #         else:
            #             # Convert to array once so broadcasting works and the dtypes match
            #             return grad * np.asarray(mask)
            #     grads = apply_mask(grads, trainable_mask)
            #     # print_cust(f"train_epochs_angle_param_adam, masked grads: {grads}")
            # (optional) mask
            if trainable_mask is not None:
                flat_mask, _ = tree_to_list(trainable_mask)
                # print_cust(f"train_epochs_angle_param_adam, flat_mask: {flat_mask}")
                # print_cust(f"train_epochs_angle_param_adam, in trainable_mask conditional, flat_g: {flat_g}")
                flat_g = [g * m for g, m in zip(flat_g, flat_mask)]
                # print_cust(f"train_epochs_angle_param_adam, in trainable_mask conditional, after applying mask, flat_g: {flat_g}")


            # print_cust(f"train_epochs_angle_param_adam, flat_g: {flat_g}")

            # print_cust(f"train_epochs_angle_param_adam, old_params: {old_params}")
            # if trainable_mask is not None:
            #   print_cust(f"train_epochs_angle_param_adam, flat_p: {flat_p}")

            # --------- Adam step ----------
            flat_p = opt.apply_grad(flat_g, flat_p)
            # if trainable_mask is not None:
            #   print_cust(f"train_epochs_angle_param_adam, flat_p after grad application: {flat_p}")

            # --------- put tensors back ----------
            params  = rebuild(flat_p)

            # print_cust(f"train_epochs_angle_param_adam, params after grad application: {params}")
            print_cust(f"train_epochs_angle_param_adam, grad_l2_norm(grads): {grad_l2_norm(flat_g)}")

            # Flatten parameters and gradients recursively
            # flat_params, shapes = flatten_params_recursive(params)
            # flat_grad = flatten_grad_recursive(grads)

            # print_cust(f"train_epochs_angle_param_adam, np.linalg.norm(flat_grad): {np.linalg.norm(flat_grad)}")
            # print_cust(f"train_epochs_angle_param_adam, flat_params.shape: {flat_params.shape}")
            # print_cust(f"train_epochs_angle_param_adam, flat_grad.shape: {flat_grad.shape}")
            # print_cust(f"train_epochs_angle_param_adam, shapes: {shapes}")

            # Convert flat_params to a Pennylane array with requires_grad=True
            # flat_params = np.array(flat_params, requires_grad=True)

            # Update parameters using the Adam optimizer
            # updated_flat_params = opt.apply_grad(flat_grad, flat_params)
            # updated_flat_params = np.array(updated_flat_params, requires_grad=True)

            # Reconstruct the nested structure of parameters
            # params = unflatten_params_recursive(updated_flat_params, shapes, structure_template)

            if tuples_allclose(old_params, params):
              print_cust(f"train_epochs_angle_param_adam, old params and params are basically the same after gradient application")
            # print_cust(f"train_epochs_angle_param_adam, grads after grad application: {grads}")
            # print_cust(f"train_epochs_angle_param_adam, params after grad application: {params}")
            loss_batch = compute_loss_angle_param_batch(params, X_batch, y_batch, layers=layers, shots=shots, batch_size=len(X_batch), qnode=qnode)
            # print_cust(f"train_epochs_angle_param_adam, params after loss computation: {params}")
            minibatch_losses.append(loss_batch)
            print_cust(f"  Mini-batch loss: {loss_batch:.4f}")

        # At the end of the epoch, compute validation loss:
        val_loss = compute_loss_angle_param_batch(params, X_val, y_val, layers=layers, shots=shots, qnode=qnode)
        validation_losses.append(val_loss)
        print_cust(f"Epoch {epoch+1} validation loss: {val_loss:.4f}")

        # Early stopping: if the validation loss doesn't improve for 'patience' epochs, stop training.
        if val_loss < best_val_loss:
            best_val_loss = val_loss
            patience_count = 0
        else:
            patience_count += 1
            if patience_count >= patience:
                print_cust("Early stopping triggered.")
                break

    return params, minibatch_losses, validation_losses


import contextlib
import pickle
from quorus.misc_utils.count_numcli_clitype import count_numcli_clitype
from quorus.misc_utils.count_qubits_layers import count_qubits_layers
from quorus.qfl_main.run_qfl_exp_multiproc import run_qfl_experiments_parallel_multiprocess
from quorus.qgan_model_supp.imggen_funcs.latent_noise_gen import generate_latent_noise
import torch.nn as nn


def main(argv=None):
  print("Running qfl_main_test.py")

  # import torch.multiprocessing as mp
  import multiprocessing as mp

  import os, random

  from pennylane import numpy as np

  import torch

  def set_seed(seed: int, deterministic: bool = False):
    """
    Sets the seed for the current run.

    Parameters:
      seed (int): The seed to set.
      deterministic (Boolean): Boolean indicating whether or not deterministic torch algorithms should be used.

    Returns:
      Nothing. Modifies seed for the current run.
    """
    # Python + NumPy
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)                          # seeds NumPy's *global* RNG

    # PyTorch (CPU + CUDA/MPS)
    torch.manual_seed(seed)                       # seeds RNG on all devices
    # Optional (esp. multi-GPU): torch.cuda.manual_seed_all(seed)

    if deterministic:
        # Make CUDA ops deterministic where possible
        torch.use_deterministic_algorithms(True)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

  
  # Define the client configuration of interest.

  dataset_types = ["Fashion-MNIST"]

  # Specifies the ansatz type for each client. A dictionary mapping client types to a list of strings, where each string
  # determines the ansatz type used for that layer. Supported strings are "reversed_staircase", "v_shape", and "revstair_vshape".
  ansatz_types_list = [
    # {10: ["reversed_staircase"],
    #  11: ["reversed_staircase", "reversed_staircase"],
    #  12: ["reversed_staircase", "reversed_staircase", "reversed_staircase"],
    #  13: ["reversed_staircase", "reversed_staircase", "reversed_staircase", "reversed_staircase"],
    #  14: ["reversed_staircase", "reversed_staircase", "reversed_staircase", "reversed_staircase", "reversed_staircase"]},

    {10: ["v_shape"],
     11: ["v_shape", "v_shape"],
     12: ["v_shape", "v_shape", "v_shape"],
     13: ["v_shape", "v_shape", "v_shape", "v_shape"],
     14: ["v_shape", "v_shape", "v_shape", "v_shape", "v_shape"]},

    # {10: ["revstair_vshape"],
    #  11: ["revstair_vshape", "revstair_vshape"],
    #  12: ["revstair_vshape", "revstair_vshape", "revstair_vshape"],
    #  13: ["revstair_vshape", "revstair_vshape", "revstair_vshape", "revstair_vshape"],
    #  14: ["revstair_vshape", "revstair_vshape", "revstair_vshape", "revstair_vshape", "revstair_vshape"]}
  ]

  class_types_list = [
    [
      ["4", "9"],
      ["3", "4"],
      ["0", "1"]
    ],
    [
      ["2", "4"],
      ["5", "8"],
      ["1", "9"]
    ]
  ]

  random_seeds = [
    12,
    30,
    50,
    70,
    400
  ]

  # Iterate over all datasets, ansatz types, classes, and seeds to run ALL specified experiments.
  for dataset_idx, dataset_type_exper in enumerate(dataset_types):
    for qubits_and_layer_types_block_params in ansatz_types_list:
      # Get the classes associated with this dataset.
      dataset_classes = class_types_list[dataset_idx]
      print(f"dataset_classes: {dataset_classes}")
      for classes_exper in dataset_classes:
        for random_state in random_seeds:
          
          # Set the seed for this run.
          set_seed(random_state)

          # Get the client configurations for this run.
          client_config_exper_parallel = {
              10: {
                  "percentage_data": 0.20,
                  "num_clients": 1,
                  "local_epochs": 1,
                  "communication_rounds": 0
              },
              11: {
                  "percentage_data": 0.20,
                  "num_clients": 1,
                  "local_epochs": 1,
                  "communication_rounds": 0
              },
              12: {
                  "percentage_data": 0.20,
                  "num_clients": 1,
                  "local_epochs": 1,
                  "communication_rounds": 0
              },
              13: {
                  "percentage_data": 0.20,
                  "num_clients": 1,
                  "local_epochs": 1,
                  "communication_rounds": 0
              },
              14: {
                  "percentage_data": 0.20,
                  "num_clients": 1,
                  "local_epochs": 1,
                  "communication_rounds": 2
              }
              # 16: {
              #     "percentage_data": 1.0,
              #     "num_clients": 4,
              #     "local_epochs": 1,
              #     "communication_rounds": 10
              # }
          }
          # Count the number of clients per client type.
          cli_type_counts_str = count_numcli_clitype(client_config_exper_parallel)
          print(f"cli_type_counts_str: {cli_type_counts_str}")
          # Specify the number of qubits and layer types used per client type. This is used for loading in initial client parameters/data.
          qubits_layers_list = [(10, 2), (11, 4), (12, 6), (13, 8), (14, 10)]
          # Convert the above list to a string, also used for loading initial parameters/data.
          qubits_layers_str = count_qubits_layers(qubits_layers_list)
          print(f"qubits_layers_str: {qubits_layers_str}")
          # Find the maximum local epochs for all clients. This is used for logging.
          local_epochs =  max([client_config_exper_parallel[key]["local_epochs"] for key in client_config_exper_parallel])
          # Specifies the path to store the folder containnig the logs for this run.
          IMG_PATH = "."
          print(f"dataset_type_exper: {dataset_type_exper}")
          # classes_exper = ["4", "9"]
          # Specifies the total amount of data, the number of clients, and the amount of data per client.
          n_samples_exper = 3640
          n_clients_tot = 5
          datapoints_per_cli = 128
          n_train_samples_exper = n_clients_tot * datapoints_per_cli
          print(f"n_train_samples_exper: {n_train_samples_exper}")
          # Find the total number of rounds to run.
          num_total_rounds_glob = max([client_config_exper_parallel[key]["communication_rounds"] for key in client_config_exper_parallel])
          # Specify the cutoff at which testing accuracy is computed every 10 rounds, for speed.
          testacc_rd_cutoff = 101
          print(f"testacc_rd_cutoff: {testacc_rd_cutoff}")
          is_multiproc = True
          # Specify the local batch size and optimizer type.
          local_batch_size = 32
          optim_type = "adam"

          # Specify learning rate for generator/discriminator for QGAN, and only discriminator for QNN.
          lr_gen = 0.004
          lr_disc = 0.001
          # Specify the decay rate of the discriminator LR.
          lr_disc_decay = 1.0
          # Specify whether or not the optimizer state should persist across rounds.
          cont_optim_state = False

          # Specify whether or not FID is computed (for QGAN).
          compute_fid = False

          # DONE: TOMODIFY, layers: have an indicator to specify that this is the classification, layers, no QCNN case.
          # DONE: TOMODIFY, layers, currently mod: use is_qcnn in the below log_data_folder name.
          # Specify whether or not to use QCNN (i.e., variational convolutional layers + feedforward) (NOTE: this is
          # an experimental feature)
          is_qcnn = False

          # Specifiy whether or not amplitude embedding is to be used (NOTE: this is an experimental feature)
          amp_embed = False
          # Specify whether clients should share their PCA or not.
          shared_pca = True
          # Specify whether or not clients should perform local PCA.
          local_pca = True

          # Specifiy whether or not parameters in later layers should have a specifal initialization, for identity initialization (NOTE:
          # this is an experimental feature).
          # This should either be "", "posneg", or "random"
          alt_zeros_init = ""

          # Specify the quantum model type to be used.
          # Should either be "", "multirun", "ancilla_endmeas", "cheating", or "tunnel_down". Details found in README.
          multiclassifier_type = "multirun"

          # Specifies whether or not the models should be trained in parallel.
          train_models_parallel = True
          # Specifies whether or not heirarchical training should be performed.
          heirarchical_train = False

          # Specifies the type of aggregation that should be performed.
          # morepers is "aggshared" (circular averaging) or "mocked_bcast" (no aggregation). Other supported types are experimental.
          morepers = "aggshared"

          # Finds the ansatz type used in the experiments.
          ansatz_type = qubits_and_layer_types_block_params[list(qubits_and_layer_types_block_params.keys())[0]][0]
          print(f"ansatz_type: {ansatz_type}")

          # random_state = 12
          # Specify the log folder where results should be stored. This string is a function of the previous configurations, in string format.
          log_data_folder = f"{IMG_PATH}/data_logs_{dataset_type_exper}_classes_{'_'.join(classes_exper)}_n_samples_{n_samples_exper}_n_train_{n_train_samples_exper}_qfl_gen_{num_total_rounds_glob}rounds_le_{local_epochs}_bs_{local_batch_size}_opt_{optim_type}_mpd_mlt_lg_{lr_gen}_ld_{lr_disc}_dqdm_ba_sp_qcnn_{is_qcnn}_ba_sm_ae_{amp_embed}_dr_nce_mc_{multiclassifier_type}_mp_{train_models_parallel}_tclip_{morepers}_{random_state}_{ansatz_type}_ldd_{lr_disc_decay}_2_10l_nos"



          # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
          # Set the device used.
          device = "cpu"

          print(f"device: {device}")

          # NOTE: tech, I don't want to mask grads...

          # import dill as pickle

          max_workers = os.cpu_count()

          mp_ctx = mp.get_context("spawn")

          # # # # TODO: redirect stdout to see what the problem is.
          # TODO: move all the stuff into if __name__ == "__main__".
          # Make the log_data_folder, if it doesn't yet exist.
          if not os.path.exists(log_data_folder):
              os.makedirs(log_data_folder)

          print("_".join(classes_exper))
          print(f"log_data_folder: {log_data_folder}")

          # import cProfile, pstats

          # Redirect output to stdout and stderr.
          with open(f"{log_data_folder}/main_stdout.txt", "w") as fout, open(f"{log_data_folder}/main_stderr.txt", "w") as ferr:
              with contextlib.redirect_stdout(fout), contextlib.redirect_stderr(ferr):

                  print(f"qubits_and_layer_types_block_params: {qubits_and_layer_types_block_params}")
                  print(f"classes_exper: {classes_exper}")
                  print(f"random_state: {random_state}")

                  mp.set_start_method("spawn", force=True)
                  # torch.multiprocessing.set_sharing_strategy('file_system')

                  # import multiprocessing as mp

                  mp_ctx = mp.get_context("spawn")

                  print(f"max_workers: {max_workers}")

                  # Specify the betas for the generator and discriminator. For the classifier case, the discriminator betas are used.
                  gen_betas = (0.5, 0.9)
                  disc_betas = (0.9, 0.99)

                  # Run the QFL workflow.
                  data_logs_prev = None
                  initial_supp_params = None
                  # TODO, public code: have a boolean indicating whether or not to use previous data.
                  with open(f"{IMG_PATH}/initial_configs_{n_clients_tot}cli_{datapoints_per_cli}tr_{n_samples_exper - (datapoints_per_cli * n_clients_tot)}test_{qubits_layers_str}/data_logs_n_samples_{n_samples_exper}_dataset_type_{dataset_type_exper}_classes_{'_'.join(classes_exper)}_train_models_parallel_{train_models_parallel}_feature_skew_0.0_label_skew_None_local_pca_{local_pca}_shared_pca_{shared_pca}_gen_False_qcnn_{is_qcnn}_{random_state}.pkl", "rb") as file:
                    data_logs_prev = pickle.load(file)
                  print(f"data_logs_prev.keys(): {data_logs_prev.keys()}")
                  print(f"data_logs_prev['clients_data_dict'].keys(): {data_logs_prev['clients_data_dict'].keys()}")
                  print(f"data_logs_prev: {data_logs_prev}")
                  # TODO, public code: have a boolean indicating whether or not to use previous parameters.
                  with open(f"{IMG_PATH}/initial_configs_{n_clients_tot}cli_{datapoints_per_cli}tr_{n_samples_exper - (datapoints_per_cli * n_clients_tot)}test_{qubits_layers_str}/client_params_dict_n_samples_{n_samples_exper}_dataset_type_{dataset_type_exper}_classes_{'_'.join(classes_exper)}_train_models_parallel_{train_models_parallel}_feature_skew_0.0_label_skew_None_local_pca_{local_pca}_shared_pca_{shared_pca}_gen_False_qcnn_{is_qcnn}_{random_state}.pkl", "rb") as file:
                    initial_supp_params = pickle.load(file)
                  print(f"initial_supp_params.keys(): {initial_supp_params.keys()}")
                  print(f"initial_supp_params: {initial_supp_params}")

                  # Commented code that is used to trim the parameters, and thus the model size. The dimensions of the supplied parameters
                  # determine the architecture/number of layers of the model.

                  # SHARED_MODEL_NUMLAYERS = 2
                  # for cli_type in initial_supp_params:
                  #   for cli_idx, cli_params in enumerate(initial_supp_params[cli_type]):
                  #     # DONE: TOMODIFY, layers: this is for the generative case; for the nongenerative case, have the code. should be pretty straightforward; just replace block params entirely
                  #     # ^ and later, can do validation to make sure that the structure, order of the supplied block params is consistent with qubits_and_layers_to_add_block_params.
                  #     # if generative:
                  #     #   cli_gen = cli_params[5][0]
                  #     #   supp_cli_gen = initial_supp_params[cli_type][cli_idx][5][0]
                  #     #   cli_gen.load_state_dict(supp_cli_gen.state_dict())
                  #     #   cli_params[5][1].load_state_dict(initial_supp_params[cli_type][cli_idx][5][1].state_dict())
                  #     # elif not is_qcnn:
                  #     cli_params_list = list(cli_params)
                  #     # new_params_list = []
                  #     # # TOMODIFY, depthFL, HACK: heuristic used
                  #     # # NOTE: assumes that there's at least one block param in the list
                  #     # num_tot_qubits = initial_supp_params[cli_type][cli_idx][5][0].shape[1]
                  #     # print(f"cli_type: {cli_type}, cli_idx: {cli_idx}, num_tot_qubits: {num_tot_qubits}")
                  #     # for orig_tens_idx, orig_param_tens in enumerate(initial_supp_params[cli_type][cli_idx][5]):
                  #     #   new_params_list.append(orig_param_tens[:, -(num_tot_qubits - orig_tens_idx):, :])
                  #     # cli_params_list[5] = new_params_list
                  #     cli_params_list[5] = initial_supp_params[cli_type][cli_idx][5][:SHARED_MODEL_NUMLAYERS]
                  #     initial_supp_params[cli_type][cli_idx] = tuple(cli_params_list)
                  
                  # print(f"initial_supp_params, after filtering extra layers: {initial_supp_params}")
                  # print(f"initial_supp_params, after filtering extra qubits: {initial_supp_params}")

                  # Specifies whether or not PCA values should be rescaled after inverse PCA, for the QGAN case (this is experimental).
                  resc_invpca = False

                  # Specifies the interface that pennylane should use.
                  pennylane_interface = "torch"

                  # Specifies the layers that each client optimizes over (NOTE: experimental; should be kept to None.)
                  opt_layers = None

                  print(f"random_state: {random_state}")

                  # Run the QFL workflow and obtain the resulting metrics, in data_logs.
                  data_logs = run_qfl_experiments_parallel_multiprocess(client_config_exper_parallel, classes=classes_exper, n_samples=n_samples_exper, dataset_type=dataset_type_exper, agg_strategy="fedavg_circ", test_frac=((n_samples_exper - n_train_samples_exper)/n_samples_exper), val_frac=0.0, random_state=random_state, pool_in=True,
                                          local_batch_size=local_batch_size, local_lr=0.01, shots=1024, debug=True, save_pkl=True, mask_grads=True, init_client_data_dict=data_logs_prev, qubits_and_layers_to_add_block_params={10: [(10, 2)], 11: [(10, 2), (10, 1)], 12: [(10, 2), (10, 1), (10, 1)], 13: [(10, 2), (10, 1), (10, 1), (10, 1)], 14: [(10, 1), (10, 1), (10, 1), (10, 1), (10, 1)]},
                                                          train_models_parallel=train_models_parallel, same_init=True, feature_skew=0.0, label_skew=None, local_pca=local_pca, do_lda=False, feat_sel_type="top", amp_embed=amp_embed, feat_ordering="same", morepers=morepers, custom_debug=True,
                                                          shared_pca=shared_pca, heirarchical_train=heirarchical_train, generative=False, use_torch=True, fed_pca_mocked=True, lr_gen=lr_gen, lr_disc=lr_disc, noise_func=generate_latent_noise, criterion_func=nn.BCELoss,
                                                              targ_data_folder_prefix="testing_gen_imgs", gen_data_folder_prefix="qgan_gen_imgs", device=device, fid_batch_size=None, max_workers=max_workers, mp_ctx=mp_ctx, log_data_folder=log_data_folder,
                                                                          initial_supp_params=initial_supp_params, optim_type=optim_type, gen_betas=gen_betas, disc_betas=disc_betas, resc_invpca=resc_invpca, compute_fid=compute_fid, is_qcnn=is_qcnn,
                                                                          pennylane_interface=pennylane_interface, opt_layers=opt_layers, alt_zeros_init=alt_zeros_init, multiclassifier_type=multiclassifier_type, testacc_rd_cutoff=testacc_rd_cutoff,
                                                                          qubits_and_layer_types_block_params=qubits_and_layer_types_block_params, lr_disc_decay=lr_disc_decay, cont_optim_state=cont_optim_state)
                  print("running after the main() function.")

                  print(f"data_logs.keys(): {data_logs.keys()}")

                  print(f"data_logs['clients_data_dict']: {data_logs['clients_data_dict']}")

                  # Save the resulting data metrics.
                  with open(f"{log_data_folder}/result_datalogs.pkl", "wb") as file:
                      pickle.dump(data_logs, file)

if __name__ == "__main__":
  main()


from pennylane import numpy as np

"""## Standard deviation of accuracy function"""

def compute_std_acc_angle_param_batch(params, X_angles, y, layers=1, shots=1024, batch_size=None, qnode=None):
    """
    Computes the standard deviation of per-class accuracies for a multiclass classifier.

    For each class, the function computes the accuracy (i.e., the fraction of correct predictions
    for examples belonging to that class) and then calculates the standard deviation of these accuracies.

    Parameters:
        params: Parameters to be passed to the qnode.
        X_angles: Array-like, inputs to the qnode function.
        y: Array-like, true labels.
        layers: (Optional) Parameter for potential layer configuration (currently unused).
        shots: (Optional) Parameter for simulation shots (currently unused).
        batch_size: (Optional) Parameter for batch processing (currently unused).
        qnode: A function that takes (input_angles, params) and returns a vector of probabilities.

    Returns:
        std_acc: The standard deviation of the per-class accuracies.
    """
    preds = []
    # Generate predictions using the qnode
    for input_angles in X_angles:
        probs = qnode(input_angles, params)
        preds.append(np.argmax(probs))

    preds = np.array(preds)

    # Compute the accuracy for each class.
    unique_classes = np.unique(y)
    class_accuracies = []

    for cls in unique_classes:
        # Indices for the current class
        indices = np.where(y == cls)[0]
        # Compute accuracy for the current class
        if len(indices) > 0:
            acc = np.mean(preds[indices] == y[indices])
            class_accuracies.append(acc)
        else:
            # If no examples are present for a class, we can either ignore or set a default accuracy (e.g., 0.0).
            class_accuracies.append(0.0)

    # Calculate the standard deviation across the per-class accuracies.
    std_acc = np.std(class_accuracies)
    return std_acc


from pennylane import numpy as np
from quorus.logging.custom_slog import print_cust
from quorus.qgan_model_supp.models.patchquantumgenerator import PatchQuantumGenerator
import torch

"""### Circular mean parallel"""

from collections import defaultdict

"""#### Block parameters aggregation function"""

def aggregate_block_params_general(client_params_dict, clients_data_dict, math_int=np):
    """
    Aggregates block (HEA ansatz) parameters for heterogeneous clients in a general way.

    Each client's block parameters (params[5]) is a list of numpy arrays,
    each with shape (num_layers, num_qubits, 3). We ignore the order in the list and
    group arrays by their block type (defined by bp.shape[1], i.e. the number of qubits).

    Args:
      client_params_dict: dict mapping client types (e.g., model sizes) to lists of parameter tuples.
          Each tuple is (conv_params, pool_params, final_pool_param, final_params, bias_param, block_params_list).
      clients_data_dict: dict mapping client types to lists of client data tuples.
          Each client data tuple is ((X_train, y_train), (X_val, y_val)).

    For each block type:
      - Let depths be the set of depths among all clients.
      - Define common_depth = min(depths). This is the number of bottom layers shared by all clients.
      - Define extra_count = max(depths) - common_depth.
      - For each client, its contribution to the common part is its bottom common_depth layers.
      - Its extra part is the top (depth - common_depth) layers.
        We align these extra layers in a meta extra region of length extra_count by computing an offset:
            offset = extra_count - (client_depth - common_depth)
        so that clients with fewer extra layers “line up” at the bottom.
      - Both the common and extra parts are aggregated using weighted circular averaging,
        with weight given by the client’s training set size.

    Returns:
      A dict mapping block type (number of qubits) to a tuple:
         (aggregated_common, aggregated_extra)
      where aggregated_common has shape (common_depth, num_qubits, 3) and
            aggregated_extra has shape (extra_count, num_qubits, 3) (or is None if no extra layers).
    """
    # Group block arrays by block type; record weight, depth, and array.
    print_cust(f"aggregate_block_params_general, math_int: {math_int}")
    block_groups = defaultdict(list)
    for ct, clients in client_params_dict.items():
        print_cust(f"aggregate_block_params_general, ct: {ct}")
        for params, data in zip(clients, clients_data_dict[ct]):
            weight = len(data[0][0])  # number of training samples for this client
            print_cust(f"aggregate_block_params_general, weight: {weight}")
            block_list = params[5]
            if block_list is None or len(block_list) == 0:
                continue
            if isinstance(block_list[0], PatchQuantumGenerator):
              block_list = block_list[0].q_params
            # NOTE: I MAY be breaking the block param functionality because I am saying that the ORDER of the block params matters.
            # Inducing additional structure; but helps for the case where you have SAME number of qubits, but
            # different LAYERS you want to optimize over for the BP's.
            for bp_idx, bp in enumerate(block_list):
                key = bp.shape[1]  # block type is determined by the number of qubits
                block_groups[(key, bp_idx)].append((weight, bp.shape[0], bp))

    aggregated_block = {}
    for key_tuple, group in block_groups.items():
        # For this block type (e.g. 4 qubits), determine the depths.
        key = key_tuple[0]
        bp_idx_group = key_tuple[1]
        depths = [item[1] for item in group]
        common_depth = min(depths)
        max_depth = max(depths)
        extra_count = max_depth - common_depth

        # Initialize accumulators for the common part:
        # Shape: (common_depth, key, 3)
        if math_int == np:
          common_accum = math_int.zeros((common_depth, key, 3), dtype=np.complex128)
        else:
          common_accum = math_int.zeros((common_depth, key, 3), dtype=torch.complex64)
        # We use a vector weight for each common layer;
        # adding a scalar to a vector broadcasts the scalar across all positions.

        if math_int == np:
          common_weight = math_int.zeros(common_depth, dtype=np.float64)
        else:
          common_weight = math_int.zeros(common_depth, dtype=torch.float32)

        # Initialize accumulators for the extra part (if any).
        if extra_count > 0:
            if math_int == np:
              extra_accum = math_int.zeros((extra_count, key, 3), dtype=np.complex128)
              extra_weight = math_int.zeros(extra_count, dtype=np.float64)
            else:
              extra_accum = math_int.zeros((extra_count, key, 3), dtype=torch.complex64)
              extra_weight = math_int.zeros(extra_count, dtype=torch.float32)
        else:
            extra_accum = None
            extra_weight = None

        # Loop over contributions.
        for weight, depth, bp in group:
            # Common part: use the bottom common_depth layers from this client's array.
            print_cust(f"aggregate_block_params_general, weight (in group loop): {weight}")
            common_part = bp[depth - common_depth : depth, :, :]  # shape (common_depth, key, 3)
            common_accum += weight * math_int.exp(1j * common_part)
            common_weight += weight  # scalar addition broadcasts to each common layer

            # Extra part: if this client has extra layers.
            client_extra = depth - common_depth
            if client_extra > 0 and extra_count > 0:
                # Align extra part: offset = extra_count - (client_extra)
                offset = extra_count - client_extra
                extra_part = bp[0 : client_extra, :, :]  # shape (client_extra, key, 3)
                # NOTE, layers: this is weird. Offset gets larger as an idx, BUT I'm (prev line) getting FIRST client_extra layers of clients' block. can change this later I guess.
                extra_accum[offset : offset + client_extra] += weight * math_int.exp(1j * extra_part)
                extra_weight[offset : offset + client_extra] += weight

        print_cust(f"aggregate_block_params_general, common_weight: {common_weight}")
        aggregated_common = math_int.angle(common_accum / common_weight[:, None, None])
        if extra_count > 0:
            aggregated_extra = math_int.angle(extra_accum / extra_weight[:, None, None])
        else:
            aggregated_extra = None
        aggregated_block[key_tuple] = (aggregated_common, aggregated_extra)

    print_cust(f"aggregate_block_params_general, aggregated_block: {aggregated_block}")

    for test_key_tuple, (test_common, test_extra) in aggregated_block.items():
      print_cust(f"aggregate_block_params_general, test_key_tuple: {test_key_tuple}")
      test_key = test_key_tuple[0]
      test_key_bpidx = test_key_tuple[1]
      print_cust(f"aggregate_block_params_general, test_key: {test_key}")
      if test_common is not None:
        print_cust(f"aggregate_block_params_general, test_common: {test_common}")
        print_cust(f"aggregate_block_params_general, test_common.requires_grad: {test_common.requires_grad}")
        print_cust(f"aggregate_block_params_general, test_common.grad_fn: {test_common.grad_fn}")
        if test_common.grad_fn is not None:
          print_cust(f"aggregate_block_params_general, test_common.grad_fn.next_functions: {test_common.grad_fn.next_functions}")
      if test_extra is not None:
        print_cust(f"aggregate_block_params_general, test_extra: {test_extra}")
        print_cust(f"aggregate_block_params_general, test_extra.requires_grad: {test_extra.requires_grad}")
        print_cust(f"aggregate_block_params_general, test_extra.grad_fn: {test_extra.grad_fn}")
        if test_extra.grad_fn is not None:
          print_cust(f"aggregate_block_params_general, test_extra.grad_fn.next_functions: {test_extra.grad_fn.next_functions}")

    return aggregated_block


from quorus.logging.custom_slog import print_cust
from quorus.qnode_funcs.qnode_creation_funcs.create_qnode_qgan import create_qnode_qgan
from quorus.qgan_model_supp.models.patchquantumgenerator import PatchQuantumGenerator
from quorus.qgan_model_supp.models.pca_discriminator import PCADiscriminator

"""## Broadcast parameters function

#### Helper Function, Convert Block Params to Generator
"""

def convert_agg_block_params_generator(aggregated_block_params):
    # This function assumes that aggregated_block_params is meant to be converted to a PatchQuantumGenerator (list of two elements; second element is PCADiscriminator)
    # This function mutates aggregated_block_params.


    # assumed to be a sequence-like obj containing first a dict, and then a PCADiscriminator.
    # focusing on the dict.
    # effectively below code, converts the dict for generator only to a model class, and then
    # sets agg_block_params to be [gen_model, disc_model] (instead of [agg_block_params_dict, disc_model])
    generator_block_params_dict = aggregated_block_params[0]
    block_list = []
    if isinstance(generator_block_params_dict, dict):
      for block_type in sorted(generator_block_params_dict.keys()):
        agg_blocks = generator_block_params_dict[block_type]
        if isinstance(agg_blocks, tuple):
          agg_blocks = agg_blocks[0]
        # assumed to be a list of tensors
        block_list.append(agg_blocks)
    # debug
    for block in block_list:
      print_cust(f"convert_agg_block_params_generator, type(block): {type(block)}")
      print_cust(f"convert_agg_block_params_generator, block.shape: {block.shape}")

    n_qubits_gen_agg = max([block.shape[1] for block in block_list])
    # NOTE: I am using global functions here which is NOT GOOD. can replace later; I am just doing this b/c I don't want to deal with
    # weight copying mutations issues.
    qgan_qnode_agg = create_qnode_qgan(n_qubits_gen_agg)
    qubits_depth_dict_agg = {}
    for block in block_list:
      # assumed that each num_qubits of each block is unique
      qubits_depth_dict_agg[block.shape[1]] = block.shape[0]
    # device_agg = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    device_agg = "cpu"
    aggregated_generator = PatchQuantumGenerator(1, 1.0, n_qubits_gen_agg, 0, qgan_qnode_agg, qubits_depth_dict_agg, device_agg, 0, True, None)

    aggregated_generator.initialize_existing_parameters(block_list)

    aggregated_block_params[0] = aggregated_generator
    assert isinstance(aggregated_block_params[1], PCADiscriminator), f"broadcast_param_updates, when agg_block_params is NOT a dict, type(aggregated_block_params[1]): {type(aggregated_block_params[1])}"

    return aggregated_block_params


"""## Helper for Alternate Zeros Init for Dynamic QCNN Params"""

from typing import Tuple, Optional, Literal
from pennylane import numpy as np
from quorus.logging.custom_slog import print_cust
import torch

def make_paired_layers(
    shape: Tuple[int, int, int],
    backend: Literal["numpy", "torch"] = "numpy",
    *,
    seed: Optional[int] = None,
    dtype=None,                 # np.dtype or torch.dtype (optional)
    device: Optional[str] = None,  # torch device string (e.g., "cuda") if backend="torch"
    requires_grad: bool = False     # only used for torch
):
    """
    Generate an array/tensor of shape (num_layers, num_qubits, 3).

    Behavior:
      - If num_layers is a multiple of 3: build triads (params, 0, -params) per group of 3.
      - Else if num_layers is a multiple of 2: use pairs as before (params, -params) per group of 2.
      - Otherwise: use the original pair logic and append leftover random layers.

    Notes:
      - When num_layers is divisible by both 2 and 3, triads take precedence.
    """
    L, Q, C = shape
    print_cust(f"make_paired_layers, shape: {shape}")
    if C != 3:
        raise ValueError(f"shape[-1] must be 3, got {C}")
    if L < 0 or Q < 0:
        raise ValueError("num_layers and num_qubits must be non-negative")

    # Decide mode (triads first to give it precedence over pairs)
    use_triads = (L % 3 == 0 and L > 0)
    use_pairs = (L % 2 == 0 and not use_triads and L > 0)

    if backend == "numpy":

        # if seed is not None:
        #     np.random.seed(seed)

        if use_triads:
            # --- TRIADS: (params, 0, -params) ---
            t = L // 3
            if t > 0:
                base = np.random.randn(t, 1, Q, 3)                         # (t,1,Q,3)
                if dtype is not None:
                    base = base.astype(dtype)
                zeros = np.zeros_like(base)                                 # (t,1,Q,3)
                triad = np.concatenate([base, zeros, -base], axis=1)       # (t,3,Q,3)
                stacked = triad.reshape(3 * t, Q, 3)                        # (3t,Q,3)
            else:
                stacked = np.empty((0, Q, 3), dtype=dtype if dtype is not None else float)
            return stacked

        # Fallback to PAIRS (original behavior, including leftovers)
        k = L // 2  # number of full pairs
        if k > 0:
            base = np.random.randn(k, 1, Q, 3)                              # (k,1,Q,3)
            if dtype is not None:
                base = base.astype(dtype)
            pair = np.concatenate([base, -base], axis=1)                    # (k,2,Q,3)
            stacked = pair.reshape(2 * k, Q, 3)                             # (2k,Q,3)
        else:
            stacked = np.empty((0, Q, 3), dtype=dtype if dtype is not None else float)

        if (L % 2) == 1:  # odd layer count → add one extra layer
            extra = np.random.randn(1, Q, 3)
            if dtype is not None:
                extra = extra.astype(dtype)
            return np.concatenate([stacked, extra], axis=0)
        return stacked

    elif backend == "torch":
        # import torch

        # if seed is not None:
        #     torch.manual_seed(seed)

        final_dtype = dtype if dtype is not None else torch.get_default_dtype()

        if use_triads:
            # --- TRIADS: (params, 0, -params) ---
            t = L // 3
            if t > 0:
                base = torch.randn((t, 1, Q, 3), dtype=final_dtype, device=device,
                                   requires_grad=requires_grad)             # (t,1,Q,3)
                zeros = torch.zeros_like(base)                               # (t,1,Q,3)
                triad = torch.cat([base, zeros, -base], dim=1)              # (t,3,Q,3)
                stacked = triad.reshape(3 * t, Q, 3)                         # (3t,Q,3)
            else:
                stacked = torch.empty((0, Q, 3), dtype=final_dtype, device=device)
                if requires_grad:
                    stacked.requires_grad_()
            return stacked

        # Fallback to PAIRS (original behavior, including leftovers)
        k = L // 2
        if k > 0:
            base = torch.randn((k, 1, Q, 3), dtype=final_dtype, device=device,
                               requires_grad=requires_grad)                  # (k,1,Q,3)
            pair = torch.cat([base, -base], dim=1)                           # (k,2,Q,3)
            stacked = pair.reshape(2 * k, Q, 3)                              # (2k,Q,3)
        else:
            stacked = torch.empty((0, Q, 3), dtype=final_dtype, device=device)
            if requires_grad:
                stacked.requires_grad_()

        if (L % 2) == 1:
            extra = torch.randn((1, Q, 3), dtype=final_dtype, device=device,
                                requires_grad=requires_grad)
            return torch.cat([stacked, extra], dim=0)
        return stacked

    else:
        raise ValueError('backend must be either "numpy" or "torch"')


import pennylane as qml
from functools import partial
from quorus.qnode_funcs.template_funcs.qgan_template import qgan_template

def create_qnode_qgan(n_data):
    """
    Creates a QNode of the specified size, for QGAN.

    Parameters:
      n_data: an integer representing the number of qubits for this QNode

    Returns:
      qnode: a QNode object satisfying the above constraints.
    """
    n_qubits = n_data
    dev = qml.device("default.qubit", wires=n_qubits)
    circuit = partial(qgan_template, n_qubits=n_qubits)
    # @qml.qnode(dev, interface="torch")
    # def qnode_qgan(input_noise, n_qubits, params):
    #     # params should just be a list of tensors representing the parameters to be used in the QGAN.
    #     # this qnode should not be exposed externally; should only be used in the context of PatchQuantumGenerator.
    #     print_cust(f"qnode_qgan, qml.draw: {qml.draw(QGAN_circuit)(input_noise, n_qubits, params)}")
    #     return QGAN_circuit(input_noise, n_qubits, params)
    # return qnode_qgan
    return qml.qnode(dev, interface="torch")(circuit)


from quorus.logging.custom_slog import print_cust
from quorus.quantum_circuit_funcs.utils.qcnn_code import compute_conv_layers
from quorus.parameter_initialization.qcnn_params_init.qcnn_params_init_func import init_dynamic_qcnn_params
from quorus.qgan_model_supp.models.pca_discriminator import PCADiscriminator
import copy
from quorus.qgan_model_supp.data_proc.pca_rescaler import PCARescaler
from quorus.qgan_model_supp.models.patchquantumgenerator import PatchQuantumGenerator

"""## Initialize client params function"""

def initialize_client_params(clients_config_arg, model_size=4, cur_client_params_dict=None, debug=False, qubits_and_layers_to_add_block_params=[],
                             train_models_parallel=False, n_output_qubits=1, generative=False, use_torch=False, qnode_func=None, device=None, pca_info=None, is_qcnn=True, alt_zeros_init=""):
    """
    Initializes the parameter dictionary storing the parameters for each client.

    Parameters:
      clients_config_arg: The client configuration dictionary used to determine the size and types of parameters that each client will have.
      model_size: Integer representing the shared model size (if that exists) for the subnet initialization case.
      cur_client_params_dict: A dictionary representing a client parameters dictionary, used for previous parameter initialization.
      debug: A Boolean indicating whether or not the parameters should be initialized in debug mode.
      qubits_and_layers_to_add_block_params: A dictionary mapping client sizes to a list of (n_qubits, n_layers) specifying the block parameters
      that this client contains.
      train_models_parallel: A Boolean indicating whether or not the models are trained in parallel (and thus, whether or not we should employ subnet initialization).
      n_output_qubits: An integer specifying the number of output qubits for these models.

    Returns:
      client_params_dict, a dictionary mapping client types (integers, representing the numbers of qubits that each client contains) to a list of parameters for each client
      (the i-th element of the list are the parameters for the i-th client)
    """
    # NOTE: for train_models_parallel, this function only needs to be called ONCE. There is no notion of 'expansion'; so the
    # subsequent checks for 'expansion' in this function should NOT be strictly necessary.
    client_params_dict = {}
    max_client_size = max(clients_config_arg.keys())
    # Iterate over each client type
    for client_type, cfg in clients_config_arg.items():
        # If the client type is smaller than the model size in expansion, or the number of clients is 0, then don't create
        # parameters for models of this type
        if (client_type < model_size and not train_models_parallel) or cfg["num_clients"] == 0:
            continue
        num_clients = cfg["num_clients"]
        client_params_dict[client_type] = []
        # For each client that is to be created of this type, initialize the parameters for this client.
        for client_idx in range(num_clients):
            # NOCHANGE: TOMODIFY, depthFL: inject some argument here to indicate that client_model_size should be dynamically found based on the largest qubits needed for its block params,
            # if not qcnn.
            # ^ client_model_size is never manifested for the non-QCNN case, so is OK to use client_model_size
            # as an alias to diff't client types in terms of number of layers.
            if train_models_parallel:
              client_model_size = client_type
            else:
              client_model_size = model_size
            is_expansion = (cur_client_params_dict is not None)
            # Filter block config: for each (num_qubits, num_layers) tuple, if num_qubits <= client_type, include it.
            qubits_and_layers_to_add_block_params_client = []
            if client_model_size in qubits_and_layers_to_add_block_params:
              qubits_and_layers_to_add_block_params_client = qubits_and_layers_to_add_block_params[client_model_size]
              if generative:
                print_cust(f"initialize_client_params, generative, filtering qubits_and_layers_to_add_block_params_client")
                qubits_and_layers_to_add_block_params_client_filt = []
                for qubits_and_layers_tuple in sorted(qubits_and_layers_to_add_block_params_client, key=lambda x:x[0]):
                  if qubits_and_layers_tuple[0] <= client_model_size:
                    qubits_and_layers_to_add_block_params_client_filt.append(qubits_and_layers_tuple)
                print_cust(f"initialize_client_params, generative, qubits_and_layers_to_add_block_params_client_filt: {qubits_and_layers_to_add_block_params_client_filt}")
                qubits_and_layers_to_add_block_params_client = qubits_and_layers_to_add_block_params_client_filt


            print_cust(f"initialize_client_params, client_idx: {client_idx}, client_model_size: {client_model_size}, n_output_qubits: {n_output_qubits}")
            print_cust(f"initialize_client_params, generative: {generative}")
            conv_layers = compute_conv_layers(client_model_size, n_output_qubits, generative=generative, is_qcnn=is_qcnn)

            conv_params_tuple, pool_params_tuple, final_pool_param, final_params, bias_param, block_params_list = \
                init_dynamic_qcnn_params(client_model_size, conv_layers, debug=debug, zeros_init=is_expansion,
                                         qubits_and_layers_to_add_block_params=qubits_and_layers_to_add_block_params_client, generative=generative, use_torch=use_torch, alt_zeros_init=alt_zeros_init)

            if generative:
              client_discriminator = PCADiscriminator(max_client_size, 4)
              client_qgan_qnode = qnode_func(client_model_size)
              client_qubits_depth_dict = {}
              for qubit_depth_tuple in qubits_and_layers_to_add_block_params_client:
                client_qubits_depth_dict[qubit_depth_tuple[0]] = qubit_depth_tuple[1]
              pca_info_used = copy.deepcopy(pca_info)
              client_pcarescaler = PCARescaler(*pca_info_used)
              client_generator = PatchQuantumGenerator(1, 1.0, client_model_size, 0, client_qgan_qnode, client_qubits_depth_dict, device, 0, True, client_pcarescaler)
              client_gan_models = [client_generator, client_discriminator]

            # TODO: continue here. initialize previous disc/generator.
            # If we want to pre-initialize the parameters for this client, do so.
            if cur_client_params_dict is not None and not train_models_parallel:
                init_trained_params = copy.deepcopy(cur_client_params_dict[client_type][client_idx])
                new_conv_params = list(conv_params_tuple)
                new_conv_params[1:] = list(init_trained_params[0])
                conv_params_tuple = tuple(new_conv_params)
                new_pool_params_tuple = list(pool_params_tuple)
                new_pool_params_tuple[1:] = list(init_trained_params[1])
                pool_params_tuple = tuple(new_pool_params_tuple)
                final_pool_param = init_trained_params[2]
                final_params = init_trained_params[3]
                bias_param = init_trained_params[4]
                # block_params_list = init_trained_params[5]
                init_block_params_list = init_trained_params[5]
                if not generative:
                  for init_block_param_idx, block_params in enumerate(init_block_params_list):
                    block_params_shape = block_params.shape
                    print_cust(f"initialize_client_params, block_params_shape: {block_params_shape}")
                    num_layers_blockparams = block_params_shape[0]
                    num_qubits_blockparams = block_params_shape[1]
                    existing_block_params = block_params_list[init_block_param_idx]
                    if existing_block_params.shape == block_params_shape:
                      block_params_list[init_block_param_idx] = block_params
                  print_cust(f"initialize_client_params, block_params_list: {block_params_list}")
                else:
                  # first time the function is called, we assume we get a block params list. but that is effectively ignored and replaced with our gen and disc.
                  # next time this function is called, I can assume that I'll get passed in a generator and discriminator model. so assume that's the format,
                  # if cur_client_params_dict is not None.
                  print_cust(f"initialize_client_params, generative, init_block_params_list: {init_block_params_list}")
                  existing_generator, existing_discriminator = init_block_params_list[0], init_block_params_list[1]
                  client_discriminator = existing_discriminator
                  client_generator = PatchQuantumGenerator(1, 0.0, client_model_size, 0, client_qgan_qnode, client_qubits_depth_dict, device, 0, True, client_pcarescaler)
                  client_generator.initialize_existing_parameters(existing_generator.q_params)
                  client_gan_models = [client_generator, client_discriminator]

            if generative:
              block_params_list = client_gan_models
              print_cust(f"initialize_client_params, generative, block_params_list: {block_params_list}")

            client_params = (conv_params_tuple, pool_params_tuple, final_pool_param, final_params, bias_param, block_params_list)
            client_params_dict[client_type].append(client_params)
    return client_params_dict


from quorus.logging.custom_slog import print_cust
import torch

"""### QGAN Expectations Helper"""

def get_expectations(noise, qnode, n_qubits_func, block_params_list):
    # Non-linear Transform
    ret_expvals = qnode(noise, block_params_list)
    # print_cust(f"get_expectations, type(ret_expvals): {type(ret_expvals)}")
    # print_cust(f"get_expectations, type(weights): {type(weights)}")
    # print_cust(f"get_expectations, len(weights): {len(weights)}")
    # for block_param in block_params_list:
    #     print_cust(f"get_expectations, block_param.shape: {block_param.shape}")
    ret_expvals_sum = sum(ret_expvals)
    print_cust(f"get_expectations, type(ret_expvals_sum): {type(ret_expvals_sum)}")
    print_cust(f"get_expectations, ret_expvals_sum: {ret_expvals_sum}")
    # weights.zero_grad()
    # ret_expvals_sum.backward()
    # print_cust(f"get_expectations, weights.grad: {weights.grad}")
    # print_cust(f"get_expectations, np.linalg.norm(weights.grad): {np.linalg.norm(weights.grad)}")
    # grad_fn_debug = qml.grad(qnode, argnum=0)
    # grads_test = grad_fn_debug(weights, n_qubits_circ=n_qubits_func, qubit_depth_dict=qubit_depth_dict, alpha=alpha, ret_exp=True)
    # print_cust(f"get_expectations, grads_test: {grads_test}")
    # print_cust(f"get_expectations, np.linalg.norm(grads_test): {np.linalg.norm(grads_test)}")
    # print_cust(f"partial_measure, probs: {probs}")
    # print_cust(f"partial_measure, torch.sum(probs): {torch.sum(probs)}")
    # probsgiven0 = probs[: (2 ** (n_qubits_func - n_a_qubits_func))]
    # # print_cust(f"partial_measure, probsgiven0: {probsgiven0}")
    # # print_cust(f"partial_measure, torch.sum(probsgiven0): {torch.sum(probsgiven0)}")
    # probsgiven0 /= torch.sum(probsgiven0)
    # # print_cust(f"partial_measure, torch.max(probsgiven0): {torch.max(probsgiven0)}")

    # # Post-Processing
    # probsgiven = probsgiven0 / torch.max(probsgiven0)
    # print_cust(f"partial_measure, probsgiven: {probsgiven}")
    torch_ret_expvals = torch.stack(ret_expvals)
    print_cust(f"get_expectations, type(torch_ret_expvals): {type(torch_ret_expvals)}")
    print_cust(f"get_expectations, torch_ret_expvals.shape: {torch_ret_expvals.shape}")
    return torch_ret_expvals


import torch
from pennylane import numpy as np

"""## Convert Data to Tensors"""

def convert_data_to_lib(clients_data_dict, math_int=np):
  """
  Function that converts the data dictionariy for each client into the specified math interface.

  Parameters:
      clients_data_dict (dict): A dictionary containing the data for each client.
      math_int (module): Either np or torch.

  Returns:
      clients_data_dict, where all data is converted to the types in math_int. Note that clients_data_dict is
      also mutated.
  """
  for client_type, clients_data in clients_data_dict.items():
    for data in clients_data:
      train_data = data[0][0]
      train_labels = data[0][1]
      # NOTE: if I'm getting datatype issues, I can change this to float32 (b/c converting from np arrs, these are float64.)
      if math_int == torch:
        data[0][0] = torch.tensor(train_data, dtype=torch.float32)
        data[0][1] = torch.tensor(train_labels, dtype=torch.float32)
      val_data = data[1][0]
      val_labels = data[1][1]
      if math_int == torch:
        data[1][0] = torch.tensor(val_data, dtype=torch.float32)
        data[1][1] = torch.tensor(val_labels, dtype=torch.float32)
  return clients_data_dict


from quorus.logging.custom_slog import print_cust
from quorus.quantum_circuit_funcs.utils.qcnn_code import compute_reduced_qubits

"""## Feature List Expansion Data Function"""

def create_feat_list_expansion_data(n_qubits, conv_layers, expand=False, pool_in=False, min_qubits_noexpand=4, train_models_parallel=False, feat_sel_type="top"):
    """
    Create the feature list (ordering of features on qubits) and expansion data (a list of (encoded_angles, reversed_indices) for
    use in identity initialization) for the specified model size and convolutional layer count.

    Parameters:
      n_qubits: Integer representing the number of qubits in the model.
      conv_layers: Integer representing the number of convolutional layers in the model.
      expand: Boolean representing whether or not this model is an expansion of a smaller model.
      pool_in: Boolean indicating whether or not the qubits should be measured/compressed inwards.
      min_qubits_noexpand: Integer representing the minimum number of qubits for which no expansion shoudl take place, across all input models.
      train_models_parallel: Boolean indicating whether or not the models are being trained in parallel (in this case, no identity initialization takes place).
      feat_sel_type: String representing the order in which the features should be chosen.

    Returns:
      feature_list: List of integers representing the ordering of the features on the qubits
      expansion_data: List of (encoded_angle_indices, reversed_indices) for specifying what input data should be initialized to the identity.
    """
    # Store the current feature list, the total number of features, as well as the remaining features.
    cur_feature_list = []
    expansion_data = []
    total_num_feats = compute_reduced_qubits(n_qubits, 0)
    remaining_feats_list = list(range(total_num_feats))

    # num_feats = n_qubits

    # TODO: check/do this only for pool_in
    # For each layer in the convolutional layer, starting from the smallest layer and increasing,
    # compute the feature list and "expand" the layers.
    for layer in range(conv_layers, -1, -1):
        num_feats = compute_reduced_qubits(n_qubits, layer)
        # No expansion data if the number of qubits is less than the minimum number of qubits for expansion.
        if num_feats < min_qubits_noexpand:
          expansion_data.append([])
          continue
        # If this is the "base" number of qubits for expansion, then initialize the feature list and expansion data.
        if num_feats == min_qubits_noexpand:
          # Either take the top features, or the top and lowest features.
          if feat_sel_type == "top":
            cur_feature_list = list(range(num_feats))
          elif feat_sel_type == "toplow":
            half_feats = num_feats // 2
            # cur_feature_list.extend(all_feats[:half_feats])
            # cur_feature_list.extend(all_feats[-half_feats:])
            cur_feature_list = remaining_feats_list[:half_feats] + remaining_feats_list[-half_feats:]
            remaining_feats_list = sorted(set(remaining_feats_list) - set(cur_feature_list))
          expansion_data.append([])
        else:
          # If this is a convolutional layer that is an expansion of a smaller layer, then for the feature list,
          # insert features to minimize the distance of the remaining qubits that are not pooled.
          # Store the order of features and qubits that were added in the expansion data.
          print_cust(f"experiment_dynamic_QCNN, cur_feature_list: {cur_feature_list}")
          insertion_indices = []
          cur_half_feats = len(cur_feature_list) // 2
          for cur_feat_idx in range(len(cur_feature_list)):
            if cur_feat_idx < cur_half_feats:
              insertion_indices.append(cur_feat_idx * 2)
            else:
              insertion_indices.append(cur_feat_idx * 2 + 1)
          print_cust(f"experiment_dynamic_QCNN, insertion_indices: {insertion_indices}")
          # Either take the top, or the top and lowest features.
          if feat_sel_type == "top":
            new_feat_idxs = list(range(len(cur_feature_list), num_feats))
          elif feat_sel_type == "toplow":
            # all_feat_idxs = sorted(set(range(num_feats)) - set(cur_feature_list))
            num_feats_to_select = num_feats - len(cur_feature_list)
            half_feats_to_select = num_feats_to_select // 2
            new_feat_idxs = remaining_feats_list[:half_feats_to_select] + remaining_feats_list[-half_feats_to_select:]
            remaining_feats_list = sorted(set(remaining_feats_list) - set(new_feat_idxs))
          print_cust(f"experiment_dynamic_QCNN, new_feat_idxs: {new_feat_idxs}")
          for new_feat_idx_idx in range(len(new_feat_idxs)):
            new_feat = new_feat_idxs[new_feat_idx_idx]
            insertion_idx = insertion_indices[new_feat_idx_idx]
            print_cust(f"experiment_dynamic_QCNN, new_feat: {new_feat}")
            print_cust(f"experiment_dynamic_QCNN, insertion_idx: {insertion_idx}")
            cur_feature_list.insert(insertion_idx, new_feat)
            print_cust(f"experiment_dynamic_QCNN, cur_feature_list: {cur_feature_list}")

          # Only add expansion data if we are expanding and want to have an identity iniialization.
          if expand == True:
            expansion_data.append([list(cur_feature_list), insertion_indices])
          else:
            expansion_data.append([])

    feature_list = list(cur_feature_list)

    # If models are trained in parallel, get rid of the expansion data because we do not have/want an identity initialization.
    if train_models_parallel == True:
      new_expansion_data = []
      for layer in range(conv_layers, -1, -1):
        new_expansion_data.append([])
      expansion_data = new_expansion_data

    expansion_data.reverse()

    print_cust(f"create_feat_list_expansion_data, feature_list: {feature_list}")
    print_cust(f"create_feat_list_expansion_data, expansion_data: {expansion_data}")

    return feature_list, expansion_data


import copy
from typing import Dict, Any

import numpy as np
import torch
import torch.nn as nn

from quorus.factory_funcs.patchquantumgen_factory import build_patchquantumgen
from quorus.factory_funcs.pca_discriminator_factory import build_pca_discriminator
from quorus.factory_funcs.variationalquantumclassif_factory import build_variationalquantumclassifier
from quorus.logging.custom_slog import print_cust
from quorus.metrics_funcs.agg_metrics_func.agg_metrics_func_eval import compute_metrics_angle_param_batch
from quorus.misc_utils.create_feat_list_expansion_data import create_feat_list_expansion_data
from quorus.training_funcs.pennylane_training.train_mod_pennylane import train_epochs_angle_param_adam
from quorus.training_funcs.torch_training.qgan_training.qgan_training_funcs import train_models
from quorus.training_funcs.torch_training.varquantumclassifier_training.train_models_classifier import train_models_classifier

"""### QFL Experiment Train Single Client Helper Function"""

import contextlib

def _train_single_client(job: Dict[str, Any]) -> Dict[str, Any]:
    """
    Worker executed in a separate process.

    Returns a dict with:
        client_type, client_idx,
        trained_params, minibatch_losses, validation_losses,
        train_acc, train_acc_stdev, train_acc_topk,
        test_acc,  test_acc_stdev,  test_acc_topk, test_loss
    """
    # --- unpack ------------------------------------------------------------
    # NOTE, layers, for client_params_indiv: it is ALL OF THE CLIENT PARAMS. list of block params is also there.
    (client_type, client_idx, client_params_indiv,
     client_train_data, client_val_data, testing_data,
     cur_model_size, min_size_clients, pool_in, feat_sel_type,
     train_models_parallel, amp_embed, shots, local_batch_size,
     local_lr, num_local_epochs, conv_layers, feat_ordering, classes,
     qnode_builder, num_total_rounds, round_num, grad_mask, generative, lr_gen, lr_disc,
     noise_func, criterion_func, log_data_folder, device, client_optims_indiv, optim_type,
     gen_betas, disc_betas, use_torch, pennylane_interface, opt_layers, layer_types_list,
     loss_type, lr_disc_decay, cont_optim_state) = (job[k] for k in
         ("client_type","client_idx","client_params_indiv",
          "client_train_data","client_val_data","testing_data",
          "cur_model_size","min_size_clients","pool_in","feat_sel_type",
          "train_models_parallel","amp_embed","shots","local_batch_size",
          "local_lr","num_local_epochs","conv_layers","feat_ordering","classes",
          "qnode_builder","num_total_rounds","round_num","grad_mask","generative",
          "lr_gen","lr_disc","noise_func","criterion_func","log_data_folder","device",
          "client_optims_indiv","optim_type","gen_betas","disc_betas","use_torch","pennylane_interface","opt_layers",
          "layer_types_list", "loss_type", "lr_disc_decay", "cont_optim_state"))

    with open(f"{log_data_folder}/cli_type_{client_type}_client_idx_{client_idx}_round_num_{round_num}_stdout.txt", "w") as fout_loc, open(f"{log_data_folder}/cli_type_{client_type}_client_idx_{client_idx}_round_num_{round_num}_stderr.txt", "w") as ferr_loc:
      with contextlib.redirect_stdout(fout_loc), contextlib.redirect_stderr(ferr_loc):

        print_cust(f"_train_single_client, opt_layers: {opt_layers}")

        print_cust(f"_train_single_client, device: {device}")

        print_cust(f"_train_single_client, qnode_builder: {qnode_builder}")

        print_cust(f"_train_single_client, lr_gen: {lr_gen}, lr_disc: {lr_disc}")

        # TODO: won't work for nongenerative case, but not going to worry about that for now.
        print_cust(f"_train_single_client, client_optims_indiv: {client_optims_indiv}")

        print_cust(f"_train_single_client, optim_type: {optim_type}")

        print_cust(f"_train_single_client, gen_betas: {gen_betas}, disc_betas: {disc_betas}")

        print_cust(f"_train_single_client, grad_mask: {grad_mask}, generative: {generative}")

        print_cust(f"_train_single_client, client_train_data: {client_train_data}")

        print_cust(f"_train_single_client, conv_layers: {conv_layers}")

        print_cust(f"_train_single_client, round_num: {round_num}, lr_disc_decay: {lr_disc_decay}")

        print_cust(f"_train_single_client, cont_optim_state: {cont_optim_state}")

        print_cust(f"_train_single_client, feat_ordering: {feat_ordering}, num_total_rounds: {num_total_rounds}")

        if not generative and conv_layers == 0:
          # NOTE, layers: override cur_model_size
          num_qubits_bps = []
          for block_param in client_params_indiv[5]:
            num_qubits_bps.append(block_param.shape[1])

          print_cust(f"_train_single_client, num_qubits_bps: {num_qubits_bps}")

          # NOTE, layers: this is a logical override.
          cur_model_size = max(num_qubits_bps)

          print_cust(f"_train_single_client, not generative and conv_layers == 0, cur_model_size: {cur_model_size}")

        # ----------------------------------------------------------------------
        # build feature list + QNode locally (nothing un‑pickle‑able is sent in)
        # DONE (ish): TOMODIFY, Layers: don't have any expansion data for the layers expansion (for now)
        # ^ TOMODIFY, LAYERS, HACK (ish): not called if conv_layers 0.
        if not generative and conv_layers > 0:
          print_cust(f"_train_single_client, calling create_feat_list_expansion_data")
          feature_list, expansion_data = create_feat_list_expansion_data(
              cur_model_size, conv_layers, expand=cur_model_size>min_size_clients,
              pool_in=pool_in, min_qubits_noexpand=min_size_clients,
              train_models_parallel=train_models_parallel,
              feat_sel_type=feat_sel_type)
        else:
          feature_list, expansion_data = list(range(cur_model_size)), []
        print_cust(f"_train_single_client, feature_list: {feature_list}, expansion_data: {expansion_data}")
        if not generative and conv_layers > 0:
          # DONE: TOMODIFY, layers: inject pennylane_interface here.
          qnode = qnode_builder(cur_model_size, conv_layers,
                                expansion_data, n_classes=len(classes), pennylane_interface=pennylane_interface)
        # DONE: TOMODIFY, layers: have another option for creating a quantumclassifer using the builder, *BUT*
        # note that the input now is the list of tensors representing the variational parameters.
        # DONE: TOMODIFY, layers: inject pennylane_interface here.
        elif not generative:
          print_cust(f"_train_single_client, conv_layers should be 0, conv_layers: {conv_layers}")

          # NOCHANGE: TOMODIFY, depthFL: add an arg for number of classifiers to create.
          # ^ TOMODIFY, depthFL: have some kind of argument to CONTROL the number/amount of classifiers I want to create.
          # classifier_data_comps = [cur_model_size, conv_layers, expansion_data, len(classes), pennylane_interface, layer_types_list, device]
          # DONE: BUG, depthFL: classifier_data_comps needs to include device somehow.... but doesn't work if it's now a dictionary.. probably, change build_vqc
          # function.
          classifier_data_comps = {
            "n_data": cur_model_size,
            "conv_layers": conv_layers,
            "expansion_data": expansion_data,
            "n_classes": len(classes),
            "pennylane_interface": pennylane_interface,
            "layer_types_list": layer_types_list,
            "device": device
          }
          cur_classifier_params = client_params_indiv
          print_cust(f"_train_single_client, cur_classifier_params: {cur_classifier_params}")
          cli_classifier_loc = build_variationalquantumclassifier(classifier_data_comps, cur_classifier_params, qnode_builder)
          print_cust(f"_train_single_client, cli_classifier_loc: {cli_classifier_loc}")
        else:
          (gen_state_dict, gen_metadata) = client_params_indiv[5][0]
          (disc_state_dict, disc_metadata) = client_params_indiv[5][1]
          print_cust(f"_train_single_client, gen_state_dict: {gen_state_dict}, gen_metadata: {gen_metadata}")
          print_cust(f"_train_single_client, disc_state_dict: {disc_state_dict}, disc_metadata: {disc_metadata}")
          cli_generator_loc = build_patchquantumgen(gen_metadata, gen_state_dict, qnode_builder)
          cli_disc_loc = build_pca_discriminator(disc_metadata, disc_state_dict)
          print_cust(f"_train_single_client, cli_generator_loc: {cli_generator_loc}")
          print_cust(f"_train_single_client, cli_disc_loc: {cli_disc_loc}")




        # ----------------------------------------------------------------------
        # slice / reorder data --------------------------------------------------
        if amp_embed:
            # TOMODIFY, layers, amp embed: comment this out. hopefully, don't need to do this; just do vanilla ampencode.
            # TOMODIFY, layers: what I should do is have a condition; if feature_list is just the range of qubits, then
            # don't do anything. Otherwise, call reorder_amplitude_data.
            print_cust(f"_train_single_client, amp_embed is True")
            # X_train = reorder_amplitude_data(client_train_data[0], feature_list)
            # X_val   = reorder_amplitude_data(client_val_data[0],  feature_list)
            # X_test  = reorder_amplitude_data(testing_data[0],      feature_list)
            X_train = client_train_data[0]
            X_val = client_val_data[0]
            X_test = testing_data[0]
        else:
            X_train = client_train_data[0][:, feature_list]
            X_val   = client_val_data[0][:,  feature_list]
            if not generative:
              X_test  = testing_data[0][:,     feature_list]
              # DONE (NO CHANGE): TOMODIFY, layers: X_test needs to be ordered in terms of feature_list.
              # ^ I mean, variational QNN is not generative

        # ----------------------------------------------------------------------
        # local training --------------------------------------------------------
        if not generative and conv_layers > 0:
          trained_params, minibatch_losses, validation_losses = train_epochs_angle_param_adam(
              copy.deepcopy(client_params_indiv),
              X_train, client_train_data[1],
              X_val,   client_val_data[1],
              n_epochs=num_local_epochs, shots=shots,
              batch_size=local_batch_size, lr=local_lr,
              qnode=qnode, trainable_mask=grad_mask)           # grad_mask handled outside

          # ----------------------------------------------------------------------
          # metrics ---------------------------------------------------------------
          if use_torch:
            math_int = torch
          else:
            math_int = np

          print_cust(f"_train_single_client, math_int: {math_int}")

          train_acc, train_acc_stdev, train_acc_topk, _ = \
              compute_metrics_angle_param_batch(trained_params, X_train,
                                                client_train_data[1],
                                                layers=conv_layers, shots=shots,
                                                batch_size=local_batch_size,
                                                qnode=qnode, math_int=math_int)
        elif not generative:
          # DONE: TOMODIFY, layers: call a different training function
          # NOT IMPLEMENTED (should be implemented in the main func): and use the updated compute_metrics_angle_param_batch function.
          # DONE: TOMODIFY, layers: initialize my optimizer from my optimizer state_dict, as before.
          print_cust(f"_train_single_client, conv_layers should be 0, conv_layers: {conv_layers}")
          client_classifier = cli_classifier_loc
          print_cust(f"_train_single_client, client_classifier: {client_classifier}")
          curr_cli_blockparams = client_classifier.q_params
          list_params = []
          if opt_layers is None:
            list_params = [block_param_torch for block_param_torch in curr_cli_blockparams]
          else:
            for layer_idx in opt_layers:
              layer_opt_param = curr_cli_blockparams[layer_idx]
              list_params.append(layer_opt_param)

          if optim_type == "sgd":
            client_optim_classifier = torch.optim.SGD(nn.ParameterList(list_params), lr=lr_disc * (lr_disc_decay ** round_num))
          elif optim_type == "adam":
            client_optim_classifier = torch.optim.Adam(nn.ParameterList(list_params), lr=lr_disc * (lr_disc_decay ** round_num), betas=disc_betas)

          if client_optims_indiv is not None and cont_optim_state:
            # NOTE: this is a SHALLOW copy of loading state dict from previous client optimizer.
            # HACK, depthFL: loading in state dict into client_optim_classifier only in non-SGD is a hack; figure out
            # how to dynamically change LR.
            print_cust(f"_train_single_client, loading state dict for client_optim_classifier")
            client_optim_classifier.load_state_dict(client_optims_indiv[5])

          print_cust(f"_train_single_client, client_optim_classifier: {client_optim_classifier}")

          print_cust(f"_train_single_client, X_train: {X_train}")

          # TODO, 8/4, 6:38 PM: continue here
          train_metrics_dict = train_models_classifier(local_batch_size, client_classifier, client_optim_classifier, criterion_func(), X_train, client_train_data[1], device, num_local_epochs, loss_type=loss_type)

          trained_classifier_params = client_classifier.q_params

          print_cust(f"_train_single_client, trained_classifier_params: {trained_classifier_params}")

          trained_optim_classifier = client_optim_classifier.state_dict()

          print_cust(f"_train_single_client, trained_optim_classifier: {trained_optim_classifier}")

          # MODIFY, layers: call compute_metrics_angle_param_batch to get the training accuracies.

        else:
          # I mean, I could return the minibatch_losses, validation_losses, but idt I will.
          print_cust(f"_train_single_client, generative is True (QGAN training)")
          client_generator = cli_generator_loc
          print_cust(f"_train_single_client, client_generator: {client_generator}")
          client_discriminator = cli_disc_loc
          print_cust(f"_train_single_client, client_discriminator: {client_discriminator}")
          # client_optim_gen = torch.optim.SGD(nn.ParameterList([client_generator.q_params[-1]]), lr=lr_gen)
          # client_optim_disc = torch.optim.SGD(client_discriminator.parameters(), lr=lr_disc)
          if optim_type == "sgd":
            client_optim_gen = torch.optim.SGD(nn.ParameterList([client_generator.q_params[-1]]), lr=lr_gen)
            client_optim_disc = torch.optim.SGD(client_discriminator.parameters(), lr=lr_disc * (lr_disc_decay ** round_num))
          elif optim_type == "adam":
            client_optim_gen = torch.optim.Adam(nn.ParameterList([client_generator.q_params[-1]]), lr=lr_gen, betas=gen_betas)
            client_optim_disc = torch.optim.Adam(client_discriminator.parameters(), lr=lr_disc * (lr_disc_decay ** round_num), betas=disc_betas)

          if client_optims_indiv is not None:
            client_optim_gen.load_state_dict(client_optims_indiv[5][0])
            client_optim_disc.load_state_dict(client_optims_indiv[5][1])

          print_cust(f"_train_single_client, client_optim_gen: {client_optim_gen}")
          print_cust(f"_train_single_client, client_optim_disc: {client_optim_disc}")
          # maybe concatenate in the train with val data? not using val data otherwise
          # can use test_result_imgs for visualization, later, if I'd like.
          print_cust(f"_train_single_client, X_train: {X_train}")
          test_result_imgs, train_metrics_dict = train_models(client_generator.n_qubits_gen, local_batch_size, client_generator, client_discriminator, client_optim_disc, client_optim_gen,
                                          noise_func=noise_func, criterion=criterion_func(), train_data=X_train, device=device, image_size=0, compressed_img_size=0, max_num_epochs=num_local_epochs,
                                          n_qubits_small=0, gen_pcas=True, disc_img_size=None, pca_disc=True)
          print_cust(f"_train_single_client, train_metrics_dict: {train_metrics_dict}")
          trained_gen_params = client_generator.state_dict()
          trained_disc_params = client_discriminator.state_dict()

          print_cust(f"_train_single_client, trained_gen_params: {trained_gen_params}")
          print_cust(f"_train_single_client, trained_disc_params: {trained_disc_params}")

          trained_optim_gen = client_optim_gen.state_dict()
          trained_optim_disc = client_optim_disc.state_dict()

          print_cust(f"_train_single_client, trained_optim_gen: {trained_optim_gen}")
          print_cust(f"_train_single_client, trained_optim_disc: {trained_optim_disc}")

        test_acc, test_acc_stdev, test_acc_topk, test_loss = (None, None, None, None)

        # test_acc, test_acc_stdev, test_acc_topk, test_loss = \
        #     compute_metrics_angle_param_batch(trained_params, X_test,
        #                                       testing_data[1],
        #                                       layers=conv_layers, shots=shots,
        #                                       batch_size=local_batch_size,
        #                                       qnode=qnode)

        # if num_total_rounds > 10:
        #   if (round_num + 1) % 10 == 0:
        #     test_acc, test_acc_stdev, test_acc_topk, test_loss = compute_metrics_angle_param_batch(trained_params, X_test, testing_data[1], layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)
        #   else:
        #     test_acc, test_acc_stdev, test_acc_topk, test_loss = (None, None, None, None)
        # else:
        #   test_acc, test_acc_stdev, test_acc_topk, test_loss = compute_metrics_angle_param_batch(trained_params, X_test, testing_data[1], layers=conv_layers, shots=shots, batch_size=local_batch_size, qnode=qnode)
        # ----------------------------------------------------------------------
        if not generative and conv_layers > 0:
          ret_dict = dict(client_type=client_type,
                    client_idx=client_idx,
                    trained_params=trained_params,
                    minibatch_losses=minibatch_losses,
                    validation_losses=validation_losses,
                    train_acc=train_acc,
                    test_acc=test_acc,
                    test_loss=test_loss,
                    train_acc_stdev=train_acc_stdev,
                    test_acc_stdev=test_acc_stdev,
                    train_acc_topk=train_acc_topk,
                    test_acc_topk=test_acc_topk)
        elif not generative:
          # NOTE, layers: trying to match the same interface as in the generative case.
          # TOMODIFY, layers: add additional training metrics (training accuracies) here.
          ret_dict = dict(client_type=client_type,
                          client_idx=client_idx,
                          trained_disc_params=trained_classifier_params,
                          trained_optim_disc=trained_optim_classifier,
                          train_metrics_dict=train_metrics_dict)
        else:
          ret_dict = dict(client_type=client_type,
                          client_idx=client_idx,
                          trained_gen_params=trained_gen_params,
                          trained_disc_params=trained_disc_params,
                          trained_optim_gen=trained_optim_gen,
                          trained_optim_disc=trained_optim_disc,
                          train_metrics_dict=train_metrics_dict)
        # DONE: TOMODIFY, layers: return new object, with trained layers params (send back the list of block params for consistency), and send back the
        # torch optimizer state dict. (Later -- can send back a training metrics dictionary.)
        # DONE (just did the point above): TOMODIFY, layers: for minimum viable things, send back whatever metrics are necessary.

    return ret_dict


"""# Data Loading

## Load MNIST Digits Helpers
"""

from sklearn.model_selection import train_test_split
from sklearn.datasets import fetch_openml
from urllib.error import HTTPError, URLError

from pennylane import numpy as np

from quorus.logging.custom_slog import print_cust

# TOMODIFY, layers: these are globals and are ONLY here to support names for fashion. it's not exactly necessary.
_FASHION_ID_TO_NAME = {
    0: "T-shirt/top", 1: "Trouser", 2: "Pullover", 3: "Dress", 4: "Coat",
    5: "Sandal", 6: "Shirt", 7: "Sneaker", 8: "Bag", 9: "Ankle boot"
}
_FASHION_NAME_TO_ID = {v.lower(): k for k, v in _FASHION_ID_TO_NAME.items()}

def _normalize_and_flatten(X):
    X = X.astype(np.float32) / 255.0
    if X.ndim == 3:  # (N, 28, 28)
        X = X.reshape((X.shape[0], -1))
    return X

def _canon_label_list(digits_to_keep, is_fashion):
    ids = []
    for d in digits_to_keep:
        s = str(d).strip()
        if s.isdigit():
            ids.append(int(s))
        elif is_fashion and s.lower() in _FASHION_NAME_TO_ID:
            ids.append(_FASHION_NAME_TO_ID[s.lower()])
        else:
            raise ValueError(
                f"Unrecognized class '{d}'. "
                + ("Use 0–9 or Fashion-MNIST names like 'sneaker', 'bag'."
                   if is_fashion else "Use digits 0–9.")
            )
    # dedupe but keep stable order by sorting later
    return sorted(set(ids))

"""## Load MNIST Digits Function"""

def load_mnist_digits(digits_to_keep, n_samples=2000, dataset_name="mnist"):
    """
    Load MNIST or Fashion-MNIST, filter to selected classes, and return (X, y).
    Tries OpenML first; if that errors, falls back to TensorFlow's keras.datasets.

    Parameters:
      digits_to_keep: list of class identifiers. For MNIST: digits (e.g., [4, 9]).
                      For Fashion-MNIST: digits 0–9 or names (e.g., ["sneaker", "bag"]).
      n_samples: number of examples to return (stratified). If >= available, returns all.
      dataset_name: "mnist" or "fashion-mnist" (case/underscore/dash tolerant).

    Returns:
      X: float32 array of shape (N, 784), normalized to [0, 1].
      y: int array of shape (N,), labels remapped to 0..K-1 following sorted(digits_to_keep).
    """
    ds = dataset_name.replace("_", "-").lower()
    # NOTE, layers: can later add some additional conditions to allow for CIFAR-10.
    if ds in {"mnist", "mnist-784"}:
        openml_name = "mnist_784"          # OpenML dataset name
        is_fashion = False
        tf_loader_path = ("tensorflow.keras.datasets.mnist", "mnist")
    elif ds in {"fashion-mnist", "fashion mnist", "fashion"}:
        openml_name = "Fashion-MNIST"      # OpenML dataset name
        is_fashion = True
        tf_loader_path = ("tensorflow.keras.datasets.fashion_mnist", "fashion_mnist")
    else:
        raise ValueError("dataset_name must be 'mnist' or 'fashion-mnist'.")

    keep_ids = _canon_label_list(digits_to_keep, is_fashion)

    # 1) Try OpenML first
    X, y = None, None
    try:
        mnist = fetch_openml(openml_name, version=1, as_frame=False)
        X = _normalize_and_flatten(mnist.data)
        # y can be strings; coerce to int if possible, else map names for fashion
        try:
            y = mnist.target.astype(int)
        except ValueError:
            # NOTE: this should be unnecessary; Fashion-MNIST from OpenML has labels that are string representations of digits.
            if is_fashion:
                y = np.array([_FASHION_NAME_TO_ID[str(lbl).lower()] for lbl in mnist.target])
            else:
                # MNIST should be numeric; re-raise if it's not
                raise
    except (HTTPError, URLError, OSError, RuntimeError, ValueError) as _:
        # 2) Fallback to TensorFlow's Keras loader
        try:
            print_cust(f"load_mnist_digits, USING TENSORFLOW TO LOAD IN MNIST DATA (b/c fetch_openml has errors)")
            print_cust(f"load_mnist_digits, error from fetch_openml code, _: {_}")
            # lazy import to avoid TF dependency unless needed
            import importlib
            mod_name, attr = tf_loader_path
            mod = importlib.import_module(mod_name)
            (X_train, y_train), (X_test, y_test) = getattr(mod, "load_data")()
            X = _normalize_and_flatten(np.concatenate([X_train, X_test], axis=0))
            y = np.concatenate([y_train, y_test], axis=0).astype(int)
        except Exception as e:
            raise RuntimeError(
                f"Both OpenML and TensorFlow loading failed: {type(e).__name__}: {e}"
            )

    # Filter to desired classes
    mask = np.isin(y, keep_ids)
    X, y = X[mask], y[mask]

    # Remap labels to 0..K-1 in sorted order of requested classes
    mapping = {label: idx for idx, label in enumerate(keep_ids)}
    y = np.array([mapping[int(lbl)] for lbl in y], dtype=int)

    # Optional stratified subsample
    if n_samples < len(y):
        # NOTE: because random_state is hardcoded in here, then we randomly select the SAME subset of data in total which is randomly sampled from
        # to create the clients' individual data.
        X, _, y, _ = train_test_split(
            X, y, train_size=n_samples, stratify=y, random_state=42
        )

    return X, y

"""## Load CIFAR-10 Data"""

# from tensorflow.keras.datasets import cifar10

def load_cifar10(classes, n_samples=2000):
    """
    Load the CIFAR-10 dataset, filter by the specified classes,
    and return the filtered data and labels.

    Parameters:
      classes (list): List of classes to keep (either as numeric strings like ["3", "5"]
                      or as names like ["cat", "dog"]).
      n_samples (int): Number of samples to return.

    Returns:
      X (np.array): Normalized and flattened image data.
      y (np.array): Integer labels remapped according to sorted(selected_classes).
    """

    # Load the data from Keras (train and test combined)
    # NOTE: cifar10 is currently not supported because it requires tensorflow.
    # You can write your own implementation for cifar10, so long as it has a load_data() method
    # that returns the data of interest.
    (X_train, y_train), (X_test, y_test) = cifar10.load_data()
    X = np.concatenate([X_train, X_test], axis=0)
    y = np.concatenate([y_train, y_test], axis=0).flatten()

    # CIFAR-10 standard class names mapping
    cifar10_classes = {
        0: "airplane", 1: "automobile", 2: "bird", 3: "cat", 4: "deer",
        5: "dog", 6: "frog", 7: "horse", 8: "ship", 9: "truck"
    }

    # Determine if classes are specified as digit strings or names.
    try:
        if classes and classes[0].isdigit():
            selected_indices = [int(c) for c in classes]
        else:
            # Normalize to lower case for comparison
            reverse_mapping = {v.lower(): k for k, v in cifar10_classes.items()}
            selected_indices = [reverse_mapping[c.lower()] for c in classes]
    except Exception as e:
        print_cust("Error processing classes for CIFAR-10: ", e)
        raise e

    mask = np.isin(y, selected_indices)
    X, y = X[mask], y[mask]

    # Create a mapping similar to load_mnist_digits, ensuring ordering is consistent.
    mapping = {val: idx for idx, val in enumerate(sorted(selected_indices))}

    def to_int(x):
        # If x is a PyTorch tensor, use .item() to extract the value.
        return int(x.item()) if hasattr(x, "item") else int(x)

    # Convert each element in y to a standard integer before mapping.
    y = np.array([mapping[to_int(val)] for val in y])

    # Normalize to [0, 1] and flatten the images.
    X = X.astype("float32") / 255.0
    X = X.reshape(X.shape[0], -1)

    # If needed, subsample the dataset via stratified sampling.
    if n_samples < len(y):
        X, _, y, _ = train_test_split(X, y, train_size=n_samples, stratify=y, random_state=42)
    return X, y


"""## QCNN QNode Creation"""

from quorus.logging.custom_slog import print_cust
from quorus.quantum_circuit_funcs.circuit_models.qcnn_circuit import QCNN_circuit_dynamic

def qcnn_template(input_angles, params, n_qubits, expansion_data, n_classes, num_ancillas=0, layer_types_list=[], cheating=False, tunn_down=False):
  """
  Runs the quantum circuit for QCNN, given specific configurations.

  Parameters:
    input_angles: the data fed into the quantum model of interest
    params: a tuple of parameters to be called with this qnode.
    n_qubits: an integer representing the number of qubits for this QNode
    expansion_data: a list of (encoded_angles, reversed_indices) that is used for identity initialization
    n_classes: an integer representing the number of output classes for this QCNN
    num_ancillas: an integer specifying the number of ancillas uised for this qnode.
    layer_types_list: a list of strings where the string at layer i specifies the ansatz type to use in this layer.
    cheating: a Boolean indicating whether or not cheating measurments should be applied.
    tunn_down: a Boolean indicating whether or not the circuit should have less layers for later layers in the circuit.

  Returns:
    a tensor or list of tensors representing the output of running through this QNode.
  """
  # TOMODIFY, depthFL: add num_ancillas to what this function calls.
  print_cust(f"qcnn_template, input_angles: {input_angles}, params: {params}, n_qubits: {n_qubits}, expansion_data: {expansion_data}, n_classes: {n_classes}, num_ancillas: {num_ancillas}, layer_types_list: {layer_types_list}, cheating: {cheating}, tunn_down: {tunn_down}")
  # print_cust(f"qcnn_template, qml.draw: {qml.draw(QCNN_circuit_dynamic)(input_angles, params[0], params[1], params[2], params[3], n_qubits, expansion_data=expansion_data, block_params_list=params[5], n_classes=n_classes, num_ancillas=num_ancillas, layer_types_list=layer_types_list, cheating=cheating, tunn_down=tunn_down)}")
  return QCNN_circuit_dynamic(input_angles, params[0], params[1], params[2], params[3],
                              n_qubits, expansion_data=expansion_data, block_params_list=params[5], n_classes=n_classes, num_ancillas=num_ancillas, layer_types_list=layer_types_list, cheating=cheating, tunn_down=tunn_down)


from quorus.logging.custom_slog import print_cust
import copy
from quorus.qnn_classifier_model_supp.variational_quantum_classifier import VariationalQuantumClassifier

"""### VariationalQuantumClassifier Factory"""

# this has a slightly different signature; doesn't use state_dict explicitly.
def build_variationalquantumclassifier(data_comps, existing_params, qnode_builder):
  """
  Factory function for building a variational quantum classifier.

  Parameters:
    data_comps (dict): Specifies the configurations of the qnode and classifier.
    existing_params (tuple): Specifies the parameters for the classifier model.
    qnode_builder (func): A function that retruns a Callable with (input_angles, params).

  Returns:
    A VariationalQuantumClassifier object that can be used for classification.
  """
  # data_comps should be [n_data, conv_layers, expansion_data, n_classes, pennylane_interface, device]
  print_cust(f"build_variationalquantumclassifier: called")
  model_device = data_comps["device"]
  # data_comps_qnode = data_comps[:-1]
  data_comps_qnode = copy.deepcopy(data_comps)
  del data_comps_qnode["device"]
  print_cust(f"build_variationalquantumclassifier, data_comps_qnode: {data_comps_qnode}")
  created_qnode = qnode_builder(**data_comps_qnode)
  built_varquantumclassifier = VariationalQuantumClassifier(created_qnode, model_device)
  built_varquantumclassifier.initialize_existing_parameters(existing_params)

  return built_varquantumclassifier



from quorus.logging.custom_slog import print_cust
from pennylane import numpy as np
from quorus.qfl_utils.aggregation_funcs.agg_utils.convert_block_params_generator import convert_agg_block_params_generator
from quorus.qgan_model_supp.models.patchquantumgenerator import PatchQuantumGenerator
import torch

"""### Broadcast parameters function, all aggregate"""

def broadcast_param_updates(client_params_dict, aggregated_params, math_int=np):
    """

    Args:
      client_params_dict: dict mapping client types (e.g. model sizes) to lists of parameter tuples.
    aggregated_params is assumed to have the structure:
      (meta_avg_conv, meta_avg_pool, avg_final_pool_param, avg_final_params, avg_bias, aggregated_block_params)
      and contains the aggregated parameters to broadcast to the clients.

    Update the clients' parameters in client_params_dict using the aggregated_params.



      - meta_avg_conv: tuple of length meta_conv_layers; each element is a tuple (even, odd)
      - meta_avg_pool: tuple of length meta_conv_layers; each element is an array
      - avg_final_pool_param: numpy array (1,)
      - avg_final_params: numpy array of shape (1,3) or similar
      - avg_bias: numpy array (bias scalar or vector)
      - aggregated_block_params: a dict mapping block type (num_qubits) to a tuple:
            (aggregated_common, aggregated_extra)
          where aggregated_common is a numpy array of shape (common_depth, num_qubits, 3) and
          aggregated_extra is either None or a numpy array of shape (meta_extra, num_qubits, 3).

    For the convolution and pooling parameters, each client’s parameters are updated based on its
    “offset” into the meta (global) parameter set.

    For the block parameters, for each block array bp (shape (client_depth, num_qubits, 3)) in the client's
    block_params_list:
      - The bottom common_depth layers are replaced by the aggregated common block for that block type.
      - If bp has extra layers (client_depth > common_depth) and an aggregated extra block exists,
        then bp’s top extra layers are updated. They are aligned using:
              offset = meta_extra - (client_depth - common_depth)
        so that the bottom of the extra region is shared among all clients.

    The function updates client_params_dict in place and returns it.
    """

    meta_avg_conv, meta_avg_pool, avg_final_pool_param, avg_final_params, avg_bias, aggregated_block_params = aggregated_params
    meta_conv_layers = len(meta_avg_conv)
    if not isinstance(aggregated_block_params, dict):
      aggregated_block_params = convert_agg_block_params_generator(aggregated_block_params)
      print_cust(f"broadcast_param_updates, aggregated_block_params: {aggregated_block_params}")

    # TODO, 7/11: load in the generator, disc params into the model for gen, disc models.

    # Iterate over each client type and update its clients.
    for client_type in client_params_dict:
        for i, params in enumerate(client_params_dict[client_type]):
            conv_params, pool_params, final_pool_param, final_params, bias_param, block_params_list = params

            # --- Update convolution parameters ---
            client_conv_layers = len(conv_params)
            offset_conv = meta_conv_layers - client_conv_layers
            new_conv_params = []
            for l in range(client_conv_layers):
                # Replace client's conv layer l with meta conv layer at l+offset_conv.
                new_conv_params.append(meta_avg_conv[l + offset_conv])

            # --- Update pooling parameters ---
            client_pool_layers = len(pool_params)
            offset_pool = meta_conv_layers - client_pool_layers
            new_pool_params = []
            for l in range(client_pool_layers):
                new_pool_params.append(meta_avg_pool[l + offset_pool])

            # --- Update final pooling, final rotation, and bias ---
            new_final_pool_param = np.array(avg_final_pool_param)  # shallow copy
            if isinstance(avg_final_params, tuple):
              new_final_params = tuple(avg_final_params)
            else:
              new_final_params = np.array(avg_final_params)
            new_bias_param = np.array(avg_bias)

            # --- Update block (HEA ansatz) parameters ---
            new_block_params_list = []
            # For each block parameter array bp in the list:
            if not isinstance(block_params_list[0], PatchQuantumGenerator):
              with torch.no_grad():
                for bp_idx, bp in enumerate(block_params_list):
                    # bp has shape (client_depth, num_qubits, 3)
                    client_depth, num_qubits, _ = bp.shape
                    # If this block type (num_qubits) has aggregated updates:
                    if (num_qubits, bp_idx) in aggregated_block_params:
                        aggregated_common, aggregated_extra = aggregated_block_params[(num_qubits, bp_idx)]
                        common_depth = aggregated_common.shape[0]  # minimum depth shared among all clients for this block type
                        if math_int == np:
                          new_bp = bp.copy()
                        elif math_int == torch:
                          print_cust(f"broadcast_param_updates, applying .detach().clone() to bp")
                          new_bp = bp.detach().clone()
                        # Update common part: last common_depth layers of bp.
                        # NOTE: in general, I do NOT want to update the last common layers; I want to update the FIRST
                        # commmon layers. but, b/c ACROSS clients of the same type, the layers should always be the same
                        # structure, it does not matter.
                        new_bp[client_depth - common_depth : client_depth, :, :] = aggregated_common
                        # Update extra part if bp has extra layers and aggregated_extra is available.
                        client_extra = client_depth - common_depth
                        if client_extra > 0 and (aggregated_extra is not None):
                            meta_extra = aggregated_extra.shape[0]
                            # Align: offset = meta_extra - client_extra
                            offset_extra = meta_extra - client_extra
                            new_bp[0:client_extra, :, :] = aggregated_extra[offset_extra : offset_extra + client_extra]
                        new_block_params_list.append(new_bp)
                    else:
                        # No aggregated update for this block type; leave unchanged.
                        print_cust(f"broadcast_param_updates, no update for bp bp_idx: {bp_idx}, client_depth: {client_depth}, num_qubits: {num_qubits}")
                        new_block_params_list.append(bp)
            else:
              # This is the generator case
              existing_generator = block_params_list[0]
              existing_discriminator = block_params_list[1]
              with torch.no_grad():
                existing_generator.load_state_dict(aggregated_block_params[0].state_dict())
                print_cust(f"broadcast_param_updates, existing_generator: {existing_generator}")
                # print_cust(f"broadcast_param_updates, existing_generator.state_dict(): {existing_generator.state_dict()}")
                existing_discriminator.load_state_dict(aggregated_block_params[1].state_dict())
              print_cust(f"broadcast_param_updates, existing_discriminator: {existing_discriminator}")
              # print_cust(f"broadcast_param_updates, existing_discriminator.state_dict(): {existing_discriminator.state_dict()}")
              new_block_params_list = block_params_list

            # Update the client parameter tuple.
            client_params_dict[client_type][i] = (tuple(new_conv_params),
                                                    tuple(new_pool_params),
                                                    new_final_pool_param,
                                                    new_final_params,
                                                    new_bias_param,
                                                    new_block_params_list)
    return client_params_dict


def count_qubits_layers(qubits_layers_list):
  """
  Function that counts the layers for each qubit/client type.

  Parameters:
      qubits_layers_list (List[Tuple(int, int)]): A list of tuples (num_qubits, num_layers) specifying, for all clients of a particular client type,
      the number of layers that they have.

  Returns:
      A string representing the qubits and layer types for this client.
  """
  list_counts = []
  for qubits_and_layers in qubits_layers_list:
    list_counts.append(str(qubits_and_layers[0]))
    list_counts.append(str(qubits_and_layers[1]))
  return "_".join(list_counts)


import copy
from quorus.logging.custom_slog import print_cust
import torch

"""## Multi-Evaluation QCNN QNode Creation"""

def run_multiprob_qnode(input_angles, params, qnode):
  """
  Runs a multi-evaluation QNode of the specified size. Runs the QNode multiple times.
  
  Parameters:
    input_angles: the data fed into the quantum model of interest
    params: a tuple of parameters to be called with this qnode.
    qnode: A Callable with inputs (input_angles, params)

  Returns:
    output_probs_tensors, a list of tensors containing the output predictions for the input.
  """
  output_probs_list = []
  block_params_list = params[5]
  for block_param_idx in range(1, len(block_params_list) + 1):
    params_copy = list(copy.deepcopy(params))
    sub_blockparams_list = block_params_list[:block_param_idx]
    params_copy[5] = sub_blockparams_list
    params_copy = tuple(params_copy)
    print_cust(f"run_multiprob_qnode, block_param_idx: {block_param_idx}, params_copy: {params_copy}")
    output_probs = qnode(input_angles, params_copy)
    output_probs_list.append(output_probs)
  output_probs_tensors = torch.stack(output_probs_list)
  print_cust(f"run_multiprob_qnode, output_probs_tensors: {output_probs_tensors}, output_probs_tensors.shape: {output_probs_tensors.shape}")
  return output_probs_tensors


"""### Save Tensors to Folder Function"""

from PIL import Image
from quorus.logging.custom_slog import print_cust
import os
import torch
from pennylane import numpy as np

def save_tensors_to_folder(img_tensors, folder_name, input_prefix, img_ext="png"):
    img_tensors = img_tensors.detach()
    if not os.path.exists(folder_name):
        os.makedirs(folder_name)
    print_cust(f"save_tensors_to_folder, img_tensors.dtype: {img_tensors.dtype}")
    print_cust(f"save_tensors_to_folder, img_tensors.max(): {img_tensors.max()}")
    print_cust(f"save_tensors_to_folder, img_tensors.min(): {img_tensors.min()}")
    assert img_tensors.max() <= 1.0, f"save_tensors_to_folder, img_tensors.max() > 1.0: {img_tensors.max()}"
    assert img_tensors.min() >= 0.0, f"save_tensors_to_folder, img_tensors.min() < 0.0: {img_tensors.min()}"
    if not img_tensors.dtype == torch.uint8 and img_tensors.max() <= 1.0:
        print_cust(f'save_tensors_to_folder: renormalizing data')
        img_tensors = (img_tensors * 255).round()
    img_tensors = img_tensors.numpy().astype(np.uint8)
    for img_idx, img_arr in enumerate(img_tensors):
        img_obj = Image.fromarray(img_arr)
        img_obj.save(f"{folder_name}/{input_prefix}_{img_idx}.{img_ext}")


from pennylane import numpy as np
import torch
from quorus.logging.custom_slog import print_cust
import pennylane as qml

"""## Aggregated metrics function"""

def compute_metrics_angle_param_batch(params, X_angles, y, layers=1, shots=1024, batch_size=None, qnode=None, math_int=np):
    """
    Function that computes the average accuracies, standard deviation of the per-class accuracies, top-k accuracy, and average loss
    with just one call of the quantum circuit.

    Parameters:
        params: Parameters to be passed to the qnode.
        X_angles: Array-like, inputs to the qnode function.
        y: Array-like, true labels.
        layers: (Optional) Parameter for potential layer configuration (currently unused).
        shots: (Optional) Parameter for simulation shots (currently unused).
        batch_size: (Optional) Parameter for batch processing (currently unused).
        qnode: A function that takes (input_angles, params) and returns a vector of probabilities.

    Returns:
        std_acc: The standard deviation of the per-class accuracies.
    """
    if math_int == np:
      container_creator = np.array
      float_dtype = np.float64
    elif math_int == torch:
      container_creator = torch.tensor
      float_dtype = torch.float32
    print_cust(f"compute_metrics_angle_param_batch, math_int: {math_int}")
    print_cust(f"compute_metrics_angle_param_batch, container_creator: {container_creator}")
    # Compute the number of classes using the first qnode call (assumes X_angles is non-empty)
    # DONE: TOMODIFY, depthFL: this is multiple output probabilities. you need to ensemble it, and/or compute the accs for EACH classifier model.
    first_probs = qnode(X_angles[0], params)
    if len(first_probs.shape) == 2:
      # then assumed to be of shape (n_classifiers, prob_dim)
      print_cust(f"compute_metrics_angle_param_batch, len(first_probs.shape) == 2, first_probs.shape: {first_probs.shape}")
      num_classes = first_probs.shape[1]
    else:
      # should be of shape prob_dim
      print_cust(f"compute_metrics_angle_param_batch, len(first_probs.shape) != 2, first_probs.shape: {first_probs.shape}")
      num_classes = len(first_probs)


    top_k_counts = math_int.zeros(num_classes)
    total_samples = len(X_angles)

    preds = []
    losses = []

    epsilon = 1e-7

    # Loop over each input and its corresponding true label.
    # DONE: TOMODIFY, depthFL: this is multiple output probabilities. you need to ensemble it, and/or compute the accs for EACH classifier model.
    all_probs = qnode(X_angles, params)
    gen_all_probs = None
    if len(all_probs.shape) == 3:
      print_cust(f"compute_metrics_angle_param_batch, len(all_probs.shape) == 3, all_probs.shape: {all_probs.shape}")
      # should be of shape (n_classifiers, input_dim, prob_dim)
      # TOMODIFY, depthFL: BEFORE taking the mean, get the statistics for each INDIVIDUAL classifier; this could be helpful.
      gen_all_probs = all_probs
      all_probs = all_probs.mean(axis=0)
    print_cust(f"compute_metrics_angle_param_batch, all_probs.shape: {all_probs.shape}")
    # of shape input_dim, prob_dim
    for i, probs in enumerate(all_probs):
        # probs = qnode(input_angles, params)
        preds.append(math_int.argmax(probs))
        # Get indices of predictions sorted from highest to lowest probability.

        sorted_indices = math_int.argsort(-container_creator(probs))
        true_label = y[i]

        loss = -qml.math.log(probs[int(true_label)] + epsilon)
        losses.append(loss)

        # For each k from 1 to the total number of classes, check if the true label is among the top k predictions.
        for k in range(1, num_classes + 1):
            if true_label in sorted_indices[:k]:
                top_k_counts[k - 1] += 1

    # Convert counts to accuracies
    top_k_accuracies = top_k_counts / total_samples


    preds = container_creator(preds)

    # Compute the accuracy for each class.
    # CHANGED, layers: wrapped y in container_creator b/c torch.unique only takes in tensors; make sure that functionality is still OK in the numpy case.
    unique_classes = math_int.unique(container_creator(y))
    class_accuracies = []

    for cls in unique_classes:
        # Indices for the current class
        indices = math_int.where(container_creator(y) == cls)[0]
        # Compute accuracy for the current class
        if len(indices) > 0:
            acc = math_int.mean(preds[indices] == container_creator(y)[indices], dtype=float_dtype)
            class_accuracies.append(acc)
        else:
            # If no examples are present for a class, we can either ignore or set a default accuracy (e.g., 0.0).
            class_accuracies.append(0.0)

    # Calculate the standard deviation across the per-class accuracies.
    # NOTE, layers (although this is not specific to the layer classifier setup): for torch, ddof=1 default; for numpy, ddof=0 default, so will get diff stddevs. (prob should use)
    # sample stddev here.
    std_acc = math_int.std(container_creator(class_accuracies))

    avg_acc = math_int.mean(preds == container_creator(y), dtype=float_dtype)

    avg_loss = qml.math.mean(qml.math.stack(losses))

    if gen_all_probs is not None:
      avg_acc_classifiers = []
      std_acc_classifiers = []
      top_k_accuracies_classifiers = []
      avg_loss_classifiers = []
      print_cust(f"compute_metrics_angle_param_batch, gen_all_probs is not None")
      # gen_all_probs assumed to be of shape 3
      for all_probs in gen_all_probs:
        # TOMODIFY, depthFL: encapsulate this logic for computing classifier accuracies in a function.
        top_k_counts = math_int.zeros(num_classes)
        total_samples = len(X_angles)

        preds = []
        losses = []

        epsilon = 1e-7

        print_cust(f"compute_metrics_angle_param_batch, all_probs.shape: {all_probs.shape}")
        # of shape input_dim, prob_dim
        for i, probs in enumerate(all_probs):
            # probs = qnode(input_angles, params)
            preds.append(math_int.argmax(probs))
            # Get indices of predictions sorted from highest to lowest probability.

            sorted_indices = math_int.argsort(-container_creator(probs))
            true_label = y[i]

            loss = -qml.math.log(probs[int(true_label)] + epsilon)
            losses.append(loss)

            # For each k from 1 to the total number of classes, check if the true label is among the top k predictions.
            for k in range(1, num_classes + 1):
                if true_label in sorted_indices[:k]:
                    top_k_counts[k - 1] += 1

        # Convert counts to accuracies
        top_k_accuracies_subclassif = top_k_counts / total_samples


        preds = container_creator(preds)

        # Compute the accuracy for each class.
        # CHANGED, layers: wrapped y in container_creator b/c torch.unique only takes in tensors; make sure that functionality is still OK in the numpy case.
        unique_classes = math_int.unique(container_creator(y))
        class_accuracies = []

        for cls in unique_classes:
            # Indices for the current class
            indices = math_int.where(container_creator(y) == cls)[0]
            # Compute accuracy for the current class
            if len(indices) > 0:
                acc = math_int.mean(preds[indices] == container_creator(y)[indices], dtype=float_dtype)
                class_accuracies.append(acc)
            else:
                # If no examples are present for a class, we can either ignore or set a default accuracy (e.g., 0.0).
                class_accuracies.append(0.0)

        # Calculate the standard deviation across the per-class accuracies.
        # NOTE, layers (although this is not specific to the layer classifier setup): for torch, ddof=1 default; for numpy, ddof=0 default, so will get diff stddevs. (prob should use)
        # sample stddev here.
        std_acc_subclassif = math_int.std(container_creator(class_accuracies))

        avg_acc_subclassif = math_int.mean(preds == container_creator(y), dtype=float_dtype)

        avg_loss_subclassif = qml.math.mean(qml.math.stack(losses))

        avg_acc_classifiers.append(avg_acc_subclassif)
        std_acc_classifiers.append(std_acc_subclassif)
        top_k_accuracies_classifiers.append(top_k_accuracies_subclassif)
        avg_loss_classifiers.append(avg_loss_subclassif)




    if gen_all_probs is None:
      return avg_acc, std_acc, top_k_accuracies, avg_loss
    else:
      print_cust(f"compute_metrics_angle_param_batch, gen_all_probs is not None, returning")
      return avg_acc, std_acc, top_k_accuracies, avg_loss, avg_acc_classifiers, std_acc_classifiers, top_k_accuracies_classifiers, avg_loss_classifiers, gen_all_probs